from django.db import NotSupportedError, models

from .managers import EmbeddedModelManager


class EmbeddedModel(models.Model):
    objects = EmbeddedModelManager()

    class Meta:
        abstract = True

    def delete(self, *args, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be deleted.")

    def save(self, *args, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be saved.")



import datetime
from decimal import Decimal
from uuid import UUID

from bson import Decimal128
from django.core.exceptions import EmptyResultSet, FullResultSet
from django.db import NotSupportedError
from django.db.models.expressions import (
    Case,
    Col,
    CombinedExpression,
    Exists,
    ExpressionWrapper,
    F,
    NegatedExpression,
    OrderBy,
    RawSQL,
    Ref,
    ResolvedOuterRef,
    Star,
    Subquery,
    Value,
    When,
)
from django.db.models.sql import Query


def case(self, compiler, connection):
    case_parts = []
    for case in self.cases:
        case_mql = {}
        try:
            case_mql["case"] = case.as_mql(compiler, connection)
        except EmptyResultSet:
            continue
        except FullResultSet:
            default_mql = case.result.as_mql(compiler, connection)
            break
        case_mql["then"] = case.result.as_mql(compiler, connection)
        case_parts.append(case_mql)
    else:
        default_mql = self.default.as_mql(compiler, connection)
    if not case_parts:
        return default_mql
    return {
        "$switch": {
            "branches": case_parts,
            "default": default_mql,
        }
    }


def col(self, compiler, connection):  # noqa: ARG001
    # If the column is part of a subquery and belongs to one of the parent
    # queries, it will be stored for reference using $let in a $lookup stage.
    # If the query is built with `alias_cols=False`, treat the column as
    # belonging to the current collection.
    if self.alias is not None and (
        self.alias not in compiler.query.alias_refcount
        or compiler.query.alias_refcount[self.alias] == 0
    ):
        try:
            index = compiler.column_indices[self]
        except KeyError:
            index = len(compiler.column_indices)
            compiler.column_indices[self] = index
        return f"$${compiler.PARENT_FIELD_TEMPLATE.format(index)}"
    # Add the column's collection's alias for columns in joined collections.
    has_alias = self.alias and self.alias != compiler.collection_name
    prefix = f"{self.alias}." if has_alias else ""
    return f"${prefix}{self.target.column}"


def combined_expression(self, compiler, connection):
    expressions = [
        self.lhs.as_mql(compiler, connection),
        self.rhs.as_mql(compiler, connection),
    ]
    return connection.ops.combine_expression(self.connector, expressions)


def expression_wrapper(self, compiler, connection):
    return self.expression.as_mql(compiler, connection)


def f(self, compiler, connection):  # noqa: ARG001
    return f"${self.name}"


def negated_expression(self, compiler, connection):
    return {"$not": expression_wrapper(self, compiler, connection)}


def order_by(self, compiler, connection):
    return self.expression.as_mql(compiler, connection)


def query(self, compiler, connection, get_wrapping_pipeline=None):
    subquery_compiler = self.get_compiler(connection=connection)
    subquery_compiler.pre_sql_setup(with_col_aliases=False)
    field_name, expr = subquery_compiler.columns[0]
    subquery = subquery_compiler.build_query(
        subquery_compiler.columns
        if subquery_compiler.query.annotations or not subquery_compiler.query.default_cols
        else None
    )
    table_output = f"__subquery{len(compiler.subqueries)}"
    from_table = next(
        e.table_name for alias, e in self.alias_map.items() if self.alias_refcount[alias]
    )
    # To perform a subquery, a $lookup stage that escapsulates the entire
    # subquery pipeline is added. The "let" clause defines the variables
    # needed to bridge the main collection with the subquery.
    subquery.subquery_lookup = {
        "as": table_output,
        "from": from_table,
        "let": {
            compiler.PARENT_FIELD_TEMPLATE.format(i): col.as_mql(compiler, connection)
            for col, i in subquery_compiler.column_indices.items()
        },
    }
    if get_wrapping_pipeline:
        # The results from some lookups must be converted to a list of values.
        # The output is compressed with an aggregation pipeline.
        wrapping_result_pipeline = get_wrapping_pipeline(
            subquery_compiler, connection, field_name, expr
        )
        # If the subquery is a combinator, wrap the result at the end of the
        # combinator pipeline...
        if subquery.query.combinator:
            subquery.combinator_pipeline.extend(wrapping_result_pipeline)
        # ... otherwise put at the end of subquery's pipeline.
        else:
            if subquery.aggregation_pipeline is None:
                subquery.aggregation_pipeline = []
            subquery.aggregation_pipeline.extend(wrapping_result_pipeline)
        # Erase project_fields since the required value is projected above.
        subquery.project_fields = None
    compiler.subqueries.append(subquery)
    return f"${table_output}.{field_name}"


def raw_sql(self, compiler, connection):  # noqa: ARG001
    raise NotSupportedError("RawSQL is not supported on MongoDB.")


def ref(self, compiler, connection):  # noqa: ARG001
    prefix = (
        f"{self.source.alias}."
        if isinstance(self.source, Col) and self.source.alias != compiler.collection_name
        else ""
    )
    return f"${prefix}{self.refs}"


def star(self, compiler, connection):  # noqa: ARG001
    return {"$literal": True}


def subquery(self, compiler, connection, get_wrapping_pipeline=None):
    return self.query.as_mql(compiler, connection, get_wrapping_pipeline=get_wrapping_pipeline)


def exists(self, compiler, connection, get_wrapping_pipeline=None):
    try:
        lhs_mql = subquery(self, compiler, connection, get_wrapping_pipeline=get_wrapping_pipeline)
    except EmptyResultSet:
        return Value(False).as_mql(compiler, connection)
    return connection.mongo_operators["isnull"](lhs_mql, False)


def when(self, compiler, connection):
    return self.condition.as_mql(compiler, connection)


def value(self, compiler, connection):  # noqa: ARG001
    value = self.value
    if isinstance(value, int):
        # Wrap numbers in $literal to prevent ambiguity when Value appears in
        # $project.
        return {"$literal": value}
    if isinstance(value, Decimal):
        return Decimal128(value)
    if isinstance(value, datetime.datetime):
        return value
    if isinstance(value, datetime.date):
        # Turn dates into datetimes since BSON doesn't support dates.
        return datetime.datetime.combine(value, datetime.datetime.min.time())
    if isinstance(value, datetime.time):
        # Turn times into datetimes since BSON doesn't support times.
        return datetime.datetime.combine(datetime.datetime.min.date(), value)
    if isinstance(value, datetime.timedelta):
        # DurationField stores milliseconds rather than microseconds.
        return value / datetime.timedelta(milliseconds=1)
    if isinstance(value, UUID):
        return value.hex
    return value


def register_expressions():
    Case.as_mql = case
    Col.as_mql = col
    CombinedExpression.as_mql = combined_expression
    Exists.as_mql = exists
    ExpressionWrapper.as_mql = expression_wrapper
    F.as_mql = f
    NegatedExpression.as_mql = negated_expression
    OrderBy.as_mql = order_by
    Query.as_mql = query
    RawSQL.as_mql = raw_sql
    Ref.as_mql = ref
    ResolvedOuterRef.as_mql = ResolvedOuterRef.as_sql
    Star.as_mql = star
    Subquery.as_mql = subquery
    When.as_mql = when
    Value.as_mql = value



from django.test import TestCase

from .forms import AuthorForm, BookForm
from .models import Address, Author, Book, Publisher


class ModelFormTests(TestCase):
    def test_update(self):
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "NY",
            "address-zip_code": "10001",
        }
        form = AuthorForm(data, instance=author)
        self.assertTrue(form.is_valid())
        form.save()
        author.refresh_from_db()
        self.assertEqual(author.age, 51)
        self.assertEqual(author.address.city, "New York City")

    def test_some_missing_data(self):
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "NY",
            "address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["address"], ["Enter all required values."])

    def test_invalid_field_data(self):
        """A field's data (state) is too long."""
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "TOO LONG",
            "address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["address"],
            [
                "Ensure this value has at most 2 characters (it has 8).",
                "Enter all required values.",
            ],
        )

    def test_all_missing_data(self):
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "",
            "address-state": "",
            "address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["address"], ["This field is required."])

    def test_nullable_field(self):
        """A nullable EmbeddedModelField is removed if all fields are empty."""
        author = Author.objects.create(
            name="Bob",
            age=50,
            address=Address(city="NYC", state="NY", zip_code="10001"),
            billing_address=Address(city="NYC", state="NY", zip_code="10001"),
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "NY",
            "address-zip_code": "10001",
            "billing_address-po_box": "",
            "billing_address-city": "",
            "billing_address-state": "",
            "billing_address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertTrue(form.is_valid())
        form.save()
        author.refresh_from_db()
        self.assertIsNone(author.billing_address)

    def test_rendering(self):
        form = AuthorForm()
        self.assertHTMLEqual(
            str(form.fields["address"].get_bound_field(form, "address")),
            """
            <div>
                <label for="id_address-po_box">PO Box:</label>
                <input id="id_address-po_box" maxlength="50" name="address-po_box" type="text">
            </div>
            <div>
                <label for="id_address-city">City:</label>
                <input type="text" name="address-city" maxlength="20" required id="id_address-city">
            </div>
            <div>
                <label for="id_address-state">State:</label>
                <input type="text" name="address-state" maxlength="2" required
                    id="id_address-state">
            </div>
            <div>
                <label for="id_address-zip_code">Zip code:</label>
                <input type="number" name="address-zip_code" required id="id_address-zip_code">
            </div>""",
        )


class NestedFormTests(TestCase):
    def test_update(self):
        book = Book.objects.create(
            title="Learning MongoDB",
            publisher=Publisher(
                name="Random House", address=Address(city="NYC", state="NY", zip_code="10001")
            ),
        )
        data = {
            "title": "Learning MongoDB!",
            "publisher-name": "Random House!",
            "publisher-address-po_box": "",
            "publisher-address-city": "New York City",
            "publisher-address-state": "NY",
            "publisher-address-zip_code": "10001",
        }
        form = BookForm(data, instance=book)
        self.assertTrue(form.is_valid())
        form.save()
        book.refresh_from_db()
        self.assertEqual(book.title, "Learning MongoDB!")
        self.assertEqual(book.publisher.name, "Random House!")
        self.assertEqual(book.publisher.address.city, "New York City")

    def test_some_missing_data(self):
        """A required field (zip_code) is missing."""
        book = Book.objects.create(
            title="Learning MongoDB",
            publisher=Publisher(
                name="Random House", address=Address(city="NYC", state="NY", zip_code="10001")
            ),
        )
        data = {
            "title": "Learning MongoDB!",
            "publisher-name": "Random House!",
            "publisher-address-po_box": "",
            "publisher-address-city": "New York City",
            "publisher-address-state": "NY",
            "publisher-address-zip_code": "",
        }
        form = BookForm(data, instance=book)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["publisher"], ["Enter all required values."])
        self.assertHTMLEqual(
            str(form),
            """
            <div>
              <label for="id_title">Title:</label>
              <input type="text" name="title" value="Learning MongoDB!" maxlength="50"
                required id="id_title">
            </div>
            <div>
              <fieldset>
                <legend>Publisher:</legend>
                <ul class="errorlist">
                  <li>Enter all required values.</li>
                </ul>
                <div>
                  <label for="id_publisher-name">Name:</label>
                  <input type="text" name="publisher-name" value="Random House!" maxlength="50"
                    required id="id_publisher-name">
                </div>
                <div>
                  <fieldset>
                    <legend>Address:</legend>
                    <div>
                      <label for="id_publisher-address-po_box">PO Box:</label>
                      <input type="text" name="publisher-address-po_box" maxlength="50"
                        id="id_publisher-address-po_box">
                    </div>
                    <div>
                      <label for="id_publisher-address-city">City:</label>
                      <input type="text" name="publisher-address-city" value="New York City"
                        maxlength="20" required id="id_publisher-address-city">
                    </div>
                    <div>
                      <label for="id_publisher-address-state">State:</label>
                      <input type="text" name="publisher-address-state" value="NY"
                        maxlength="2" required id="id_publisher-address-state">
                    </div>
                    <div>
                      <label for="id_publisher-address-zip_code">Zip code:</label>
                      <input type="number" name="publisher-address-zip_code"
                        required id="id_publisher-address-zip_code">
                    </div>
                  </fieldset>
                </div>
              </fieldset>
            </div>""",
        )

    def test_invalid_field_data(self):
        """A field's data (state) is too long."""
        book = Book.objects.create(
            title="Learning MongoDB",
            publisher=Publisher(
                name="Random House", address=Address(city="NYC", state="NY", zip_code="10001")
            ),
        )
        data = {
            "title": "Learning MongoDB!",
            "publisher-name": "Random House!",
            "publisher-address-po_box": "",
            "publisher-address-city": "New York City",
            "publisher-address-state": "TOO LONG",
            "publisher-address-zip_code": "10001",
        }
        form = BookForm(data, instance=book)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["publisher"],
            ["Ensure this value has at most 2 characters (it has 8)."],
        )
        self.assertHTMLEqual(
            str(form),
            """
            <div>
              <label for="id_title">Title:</label>
              <input type="text" name="title" value="Learning MongoDB!"
                maxlength="50" required id="id_title">
            </div>
            <div>
              <fieldset>
                <legend>Publisher:</legend>
                <ul class="errorlist">
                  <li>Ensure this value has at most 2 characters (it has 8).</li>
                </ul>
                <div>
                  <label for="id_publisher-name">Name:</label>
                  <input type="text" name="publisher-name" value="Random House!"
                    maxlength="50" required id="id_publisher-name">
                </div>
                <div>
                  <fieldset>
                    <legend>Address:</legend>
                    <div>
                      <label for="id_publisher-address-po_box">PO Box:</label>
                      <input type="text" name="publisher-address-po_box"
                        maxlength="50" id="id_publisher-address-po_box">
                    </div>
                    <div>
                      <label for="id_publisher-address-city">City:</label>
                      <input type="text" name="publisher-address-city" value="New York City"
                        maxlength="20" required id="id_publisher-address-city">
                    </div>
                    <div>
                      <label for="id_publisher-address-state">State:</label>
                      <input type="text" name="publisher-address-state" value="TOO LONG"
                        maxlength="2" required id="id_publisher-address-state">
                    </div>
                    <div>
                      <label for="id_publisher-address-zip_code">Zip code:</label>
                      <input type="number" name="publisher-address-zip_code" value="10001"
                        required id="id_publisher-address-zip_code">
                    </div>
                  </fieldset>
                </div>
              </fieldset>
            </div>""",
        )

    def test_all_missing_data(self):
        """An embedded model with all data missing triggers a required error."""
        book = Book.objects.create(
            title="Learning MongoDB",
            publisher=Publisher(
                name="Random House", address=Address(city="NYC", state="NY", zip_code="10001")
            ),
        )
        data = {
            "title": "Learning MongoDB!",
            "publisher-name": "Random House!",
            "publisher-address-po_box": "",
            "publisher-address-city": "",
            "publisher-address-state": "",
            "publisher-address-zip_code": "",
        }
        form = BookForm(data, instance=book)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["publisher"], ["This field is required."])

    def test_rendering(self):
        form = BookForm()
        self.assertHTMLEqual(
            str(form.fields["publisher"].get_bound_field(form, "publisher")),
            """
            <div>
              <label for="id_publisher-name">Name:</label>
              <input type="text" name="publisher-name" maxlength="50"
                required id="id_publisher-name">
            </div>
            <div>
              <fieldset>
                <legend>Address:</legend>
                <div>
                  <label for="id_publisher-address-po_box">PO Box:</label>
                  <input type="text" name="publisher-address-po_box" maxlength="50"
                    id="id_publisher-address-po_box">
                </div>
                <div>
                  <label for="id_publisher-address-city">City:</label>
                  <input type="text" name="publisher-address-city" maxlength="20"
                    required id="id_publisher-address-city">
                </div>
                <div>
                  <label for="id_publisher-address-state">State:</label>
                  <input type="text" name="publisher-address-state" maxlength="2"
                    required id="id_publisher-address-state">
                </div>
                <div>
                  <label for="id_publisher-address-zip_code">Zip code:</label>
                  <input type="number" name="publisher-address-zip_code"
                    required id="id_publisher-address-zip_code">
                </div>
              </fieldset>
            </div>""",
        )



from django.db import NotSupportedError
from django.db.models import Func
from django.test import SimpleTestCase


class FuncTests(SimpleTestCase):
    def test_no_as_mql(self):
        msg = "Func() may need an as_mql() method."
        with self.assertRaisesMessage(NotSupportedError, msg):
            Func().as_mql(None, None)



from sphinx.domains.std import Cmdoption


def setup(app):
    app.add_object_type(
        directivename="django-admin",
        rolename="djadmin",
        indextemplate="pair: %s; django-admin command",
    )
    app.add_crossref_type(
        directivename="fieldlookup",
        rolename="lookup",
        indextemplate="pair: %s; field lookup type",
    )
    app.add_crossref_type(
        directivename="setting",
        rolename="setting",
        indextemplate="pair: %s; setting",
    )
    app.add_directive("django-admin-option", Cmdoption)



from django.db import models


class Number(models.Model):
    num = models.IntegerField(blank=True, null=True)

    class Meta:
        ordering = ("num",)

    def __str__(self):
        return str(self.num)



from django.db import migrations, models

import django_mongodb_backend


class Migration(migrations.Migration):
    dependencies = [
        ("model_fields_", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="integerarraydefaultmodel",
            name="field_2",
            field=django_mongodb_backend.fields.ArrayField(models.IntegerField(), default=[]),
            preserve_default=False,
        ),
    ]



__version__ = "5.1.0b3"

# Check Django compatibility before other imports which may fail if the
# wrong version of Django is installed.
from .utils import check_django_compatability, parse_uri

check_django_compatability()

from .aggregates import register_aggregates  # noqa: E402
from .expressions import register_expressions  # noqa: E402
from .fields import register_fields  # noqa: E402
from .functions import register_functions  # noqa: E402
from .indexes import register_indexes  # noqa: E402
from .lookups import register_lookups  # noqa: E402
from .query import register_nodes  # noqa: E402

__all__ = ["parse_uri"]

register_aggregates()
register_expressions()
register_fields()
register_functions()
register_indexes()
register_lookups()
register_nodes()



from django.db.models.aggregates import Aggregate, Count, StdDev, Variance
from django.db.models.expressions import Case, Value, When
from django.db.models.lookups import IsNull

from .query_utils import process_lhs

# Aggregates whose MongoDB aggregation name differ from Aggregate.function.lower().
MONGO_AGGREGATIONS = {Count: "sum"}


def aggregate(
    self,
    compiler,
    connection,
    operator=None,
    resolve_inner_expression=False,
    **extra_context,  # noqa: ARG001
):
    if self.filter:
        node = self.copy()
        node.filter = None
        source_expressions = node.get_source_expressions()
        condition = When(self.filter, then=source_expressions[0])
        node.set_source_expressions([Case(condition)] + source_expressions[1:])
    else:
        node = self
    lhs_mql = process_lhs(node, compiler, connection)
    if resolve_inner_expression:
        return lhs_mql
    operator = operator or MONGO_AGGREGATIONS.get(self.__class__, self.function.lower())
    return {f"${operator}": lhs_mql}


def count(self, compiler, connection, resolve_inner_expression=False, **extra_context):  # noqa: ARG001
    """
    When resolve_inner_expression=True, return the MQL that resolves as a
    value. This is used to count different elements, so the inner values are
    returned to be pushed into a set.
    """
    if not self.distinct or resolve_inner_expression:
        if self.filter:
            node = self.copy()
            node.filter = None
            source_expressions = node.get_source_expressions()
            condition = When(
                self.filter, then=Case(When(IsNull(source_expressions[0], False), then=Value(1)))
            )
            node.set_source_expressions([Case(condition)] + source_expressions[1:])
            inner_expression = process_lhs(node, compiler, connection)
        else:
            lhs_mql = process_lhs(self, compiler, connection)
            null_cond = {"$in": [{"$type": lhs_mql}, ["missing", "null"]]}
            inner_expression = {
                "$cond": {"if": null_cond, "then": None, "else": lhs_mql if self.distinct else 1}
            }
        if resolve_inner_expression:
            return inner_expression
        return {"$sum": inner_expression}
    # If distinct=True or resolve_inner_expression=False, sum the size of the
    # set.
    lhs_mql = process_lhs(self, compiler, connection)
    # None shouldn't be counted, so subtract 1 if it's present.
    exits_null = {"$cond": {"if": {"$in": [{"$literal": None}, lhs_mql]}, "then": -1, "else": 0}}
    return {"$add": [{"$size": lhs_mql}, exits_null]}


def stddev_variance(self, compiler, connection, **extra_context):
    if self.function.endswith("_SAMP"):
        operator = "stdDevSamp"
    elif self.function.endswith("_POP"):
        operator = "stdDevPop"
    return aggregate(self, compiler, connection, operator=operator, **extra_context)


def register_aggregates():
    Aggregate.as_mql = aggregate
    Count.as_mql = count
    StdDev.as_mql = stddev_variance
    Variance.as_mql = stddev_variance



from django.db import NotSupportedError
from django.db.models.fields.json import (
    ContainedBy,
    DataContains,
    HasAnyKeys,
    HasKey,
    HasKeyLookup,
    HasKeys,
    JSONExact,
    KeyTransform,
    KeyTransformIn,
    KeyTransformIsNull,
    KeyTransformNumericLookupMixin,
)

from ..lookups import builtin_lookup
from ..query_utils import process_lhs, process_rhs


def build_json_mql_path(lhs, key_transforms):
    # Build the MQL path using the collected key transforms.
    result = lhs
    for key in key_transforms:
        get_field = {"$getField": {"input": result, "field": key}}
        # Handle array indexing if the key is a digit. If key is something
        # like '001', it's not an array index despite isdigit() returning True.
        if key.isdigit() and str(int(key)) == key:
            result = {
                "$cond": {
                    "if": {"$isArray": result},
                    "then": {"$arrayElemAt": [result, int(key)]},
                    "else": get_field,
                }
            }
        else:
            result = get_field
    return result


def contained_by(self, compiler, connection):  # noqa: ARG001
    raise NotSupportedError("contained_by lookup is not supported on this database backend.")


def data_contains(self, compiler, connection):  # noqa: ARG001
    raise NotSupportedError("contains lookup is not supported on this database backend.")


def _has_key_predicate(path, root_column, negated=False):
    """Return MQL to check for the existence of `path`."""
    result = {
        "$and": [
            # The path must exist (i.e. not be "missing").
            {"$ne": [{"$type": path}, "missing"]},
            # If the JSONField value is None, an additional check for not null
            # is needed since $type returns null instead of "missing".
            {"$ne": [root_column, None]},
        ]
    }
    if negated:
        result = {"$not": result}
    return result


def has_key_lookup(self, compiler, connection):
    """Return MQL to check for the existence of a key."""
    rhs = self.rhs
    lhs = process_lhs(self, compiler, connection)
    if not isinstance(rhs, list | tuple):
        rhs = [rhs]
    paths = []
    # Transform any "raw" keys into KeyTransforms to allow consistent handling
    # in the code that follows.
    for key in rhs:
        rhs_json_path = key if isinstance(key, KeyTransform) else KeyTransform(key, self.lhs)
        paths.append(rhs_json_path.as_mql(compiler, connection))
    keys = []
    for path in paths:
        keys.append(_has_key_predicate(path, lhs))
    if self.mongo_operator is None:
        return keys[0]
    return {self.mongo_operator: keys}


_process_rhs = JSONExact.process_rhs


def json_exact_process_rhs(self, compiler, connection):
    """Skip JSONExact.process_rhs()'s conversion of None to "null"."""
    return (
        super(JSONExact, self).process_rhs(compiler, connection)
        if connection.vendor == "mongodb"
        else _process_rhs(self, compiler, connection)
    )


def key_transform(self, compiler, connection):
    """
    Return MQL for this KeyTransform (JSON path).

    JSON paths cannot always be represented simply as $var.key1.key2.key3 due
    to possible array types. Therefore, indexing arrays requires the use of
    `arrayElemAt`. Additionally, $cond is necessary to verify the type before
    performing the operation.
    """
    key_transforms = [self.key_name]
    previous = self.lhs
    # Collect all key transforms in order.
    while isinstance(previous, KeyTransform):
        key_transforms.insert(0, previous.key_name)
        previous = previous.lhs
    lhs_mql = previous.as_mql(compiler, connection)
    return build_json_mql_path(lhs_mql, key_transforms)


def key_transform_in(self, compiler, connection):
    """
    Return MQL to check if a JSON path exists and that its values are in the
    set of specified values (rhs).
    """
    lhs_mql = process_lhs(self, compiler, connection)
    # Traverse to the root column.
    previous = self.lhs
    while isinstance(previous, KeyTransform):
        previous = previous.lhs
    root_column = previous.as_mql(compiler, connection)
    value = process_rhs(self, compiler, connection)
    # Construct the expression to check if lhs_mql values are in rhs values.
    expr = connection.mongo_operators[self.lookup_name](lhs_mql, value)
    return {"$and": [_has_key_predicate(lhs_mql, root_column), expr]}


def key_transform_is_null(self, compiler, connection):
    """
    Return MQL to check the nullability of a key.

    If `isnull=True`, the query matches objects where the key is missing or the
    root column is null. If `isnull=False`, the query negates the result to
    match objects where the key exists.

    Reference: https://code.djangoproject.com/ticket/32252
    """
    lhs_mql = process_lhs(self, compiler, connection)
    rhs_mql = process_rhs(self, compiler, connection)
    # Get the root column.
    previous = self.lhs
    while isinstance(previous, KeyTransform):
        previous = previous.lhs
    root_column = previous.as_mql(compiler, connection)
    return _has_key_predicate(lhs_mql, root_column, negated=rhs_mql)


def key_transform_numeric_lookup_mixin(self, compiler, connection):
    """
    Return MQL to check if the field exists (i.e., is not "missing" or "null")
    and that the field matches the given numeric lookup expression.
    """
    expr = builtin_lookup(self, compiler, connection)
    lhs = process_lhs(self, compiler, connection)
    # Check if the type of lhs is not "missing" or "null".
    not_missing_or_null = {"$not": {"$in": [{"$type": lhs}, ["missing", "null"]]}}
    return {"$and": [expr, not_missing_or_null]}


def register_json_field():
    ContainedBy.as_mql = contained_by
    DataContains.as_mql = data_contains
    HasAnyKeys.mongo_operator = "$or"
    HasKey.mongo_operator = None
    HasKeyLookup.as_mql = has_key_lookup
    HasKeys.mongo_operator = "$and"
    JSONExact.process_rhs = json_exact_process_rhs
    KeyTransform.as_mql = key_transform
    KeyTransformIn.as_mql = key_transform_in
    KeyTransformIsNull.as_mql = key_transform_is_null
    KeyTransformNumericLookupMixin.as_mql = key_transform_numeric_lookup_mixin



import difflib

from django.core import checks
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.fields.related import lazy_related_operation
from django.db.models.lookups import Transform

from .. import forms


class EmbeddedModelField(models.Field):
    """Field that stores a model instance."""

    def __init__(self, embedded_model, *args, **kwargs):
        """
        `embedded_model` is the model class of the instance to be stored.
        Like other relational fields, it may also be passed as a string.
        """
        self.embedded_model = embedded_model
        super().__init__(*args, **kwargs)

    def check(self, **kwargs):
        from ..models import EmbeddedModel

        errors = super().check(**kwargs)
        if not issubclass(self.embedded_model, EmbeddedModel):
            return [
                checks.Error(
                    "Embedded models must be a subclass of "
                    "django_mongodb_backend.models.EmbeddedModel.",
                    obj=self,
                    id="django_mongodb_backend.embedded_model.E002",
                )
            ]
        for field in self.embedded_model._meta.fields:
            if field.remote_field:
                errors.append(
                    checks.Error(
                        "Embedded models cannot have relational fields "
                        f"({self.embedded_model().__class__.__name__}.{field.name} "
                        f"is a {field.__class__.__name__}).",
                        obj=self,
                        id="django_mongodb_backend.embedded_model.E001",
                    )
                )
        return errors

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("django_mongodb_backend.fields.embedded_model"):
            path = path.replace(
                "django_mongodb_backend.fields.embedded_model", "django_mongodb_backend.fields"
            )
        kwargs["embedded_model"] = self.embedded_model
        return name, path, args, kwargs

    def get_internal_type(self):
        return "EmbeddedModelField"

    def _set_model(self, model):
        """
        Resolve embedded model class once the field knows the model it belongs
        to. If __init__()'s embedded_model argument is a string, resolve it to
        the actual model class, similar to relation fields.
        """
        self._model = model
        if model is not None and isinstance(self.embedded_model, str):

            def _resolve_lookup(_, resolved_model):
                self.embedded_model = resolved_model

            lazy_related_operation(_resolve_lookup, model, self.embedded_model)

    model = property(lambda self: self._model, _set_model)

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def to_python(self, value):
        """
        Pass embedded model fields' values through each field's to_python() and
        reinstantiate the embedded instance.
        """
        if value is None:
            return None
        if not isinstance(value, dict):
            return value
        instance = self.embedded_model(
            **{
                field.attname: field.to_python(value[field.attname])
                for field in self.embedded_model._meta.fields
                if field.attname in value
            }
        )
        instance._state.adding = False
        return instance

    def get_db_prep_save(self, embedded_instance, connection):
        """
        Apply pre_save() and get_db_prep_save() of embedded instance fields and
        create the {field: value} dict to be saved.
        """
        if embedded_instance is None:
            return None
        if not isinstance(embedded_instance, self.embedded_model):
            raise TypeError(
                f"Expected instance of type {self.embedded_model!r}, not "
                f"{type(embedded_instance)!r}."
            )
        field_values = {}
        add = embedded_instance._state.adding
        for field in embedded_instance._meta.fields:
            value = field.get_db_prep_save(
                field.pre_save(embedded_instance, add), connection=connection
            )
            # Exclude unset primary keys (e.g. {'id': None}).
            if field.primary_key and value is None:
                continue
            field_values[field.attname] = value
        # This instance will exist in the database soon.
        embedded_instance._state.adding = False
        return field_values

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        field = self.embedded_model._meta.get_field(name)
        return KeyTransformFactory(name, field)

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        if self.embedded_model is None:
            return
        for field in self.embedded_model._meta.fields:
            attname = field.attname
            field.validate(getattr(value, attname), model_instance)

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": forms.EmbeddedModelField,
                "model": self.embedded_model,
                "prefix": self.name,
                **kwargs,
            }
        )


class KeyTransform(Transform):
    def __init__(self, key_name, ref_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key_name = str(key_name)
        self.ref_field = ref_field

    def get_lookup(self, name):
        return self.ref_field.get_lookup(name)

    def get_transform(self, name):
        """
        Validate that `name` is either a field of an embedded model or a
        lookup on an embedded model's field.
        """
        if transform := self.ref_field.get_transform(name):
            return transform
        suggested_lookups = difflib.get_close_matches(name, self.ref_field.get_lookups())
        if suggested_lookups:
            suggested_lookups = " or ".join(suggested_lookups)
            suggestion = f", perhaps you meant {suggested_lookups}?"
        else:
            suggestion = "."
        raise FieldDoesNotExist(
            f"Unsupported lookup '{name}' for "
            f"{self.ref_field.__class__.__name__} '{self.ref_field.name}'"
            f"{suggestion}"
        )

    def as_mql(self, compiler, connection):
        previous = self
        key_transforms = []
        while isinstance(previous, KeyTransform):
            key_transforms.insert(0, previous.key_name)
            previous = previous.lhs
        mql = previous.as_mql(compiler, connection)
        transforms = ".".join(key_transforms)
        return f"{mql}.{transforms}"

    @property
    def output_field(self):
        return self.ref_field


class KeyTransformFactory:
    def __init__(self, key_name, ref_field):
        self.key_name = key_name
        self.ref_field = ref_field

    def __call__(self, *args, **kwargs):
        return KeyTransform(self.key_name, self.ref_field, *args, **kwargs)



#!/usr/bin/env python
import os
import pathlib
import sys

test_apps = [
    "admin_changelist",
    "admin_checks",
    "admin_custom_urls",
    "admin_docs",
    "admin_filters",
    "admin_inlines",
    "admin_ordering",
    "admin_scripts",
    "admin_utils",
    "admin_views",
    "admin_widgets",
    "aggregation",
    "aggregation_regress",
    "annotations",
    "apps",
    "async",
    "auth_tests",
    "backends",
    "basic",
    "bulk_create",
    "cache",
    "check_framework",
    "constraints",
    "contenttypes_tests",
    "context_processors",
    "custom_columns",
    "custom_lookups",
    "custom_managers",
    "custom_pk",
    "datatypes",
    "dates",
    "datetimes",
    "db_functions",
    "defer",
    "defer_regress",
    "delete",
    "delete_regress",
    "empty",
    "empty_models",
    "expressions",
    "expressions_case",
    "field_defaults",
    "file_storage",
    "file_uploads",
    "fixtures",
    "fixtures_model_package",
    "fixtures_regress",
    "flatpages_tests",
    "force_insert_update",
    "foreign_object",
    "forms_tests",
    "from_db_value",
    "generic_inline_admin",
    "generic_relations",
    "generic_relations_regress",
    "generic_views",
    "get_earliest_or_latest",
    "get_object_or_404",
    "get_or_create",
    "i18n",
    "indexes",
    "inline_formsets",
    "introspection",
    "invalid_models_tests",
    "known_related_objects",
    "lookup",
    "m2m_and_m2o",
    "m2m_intermediary",
    "m2m_multiple",
    "m2m_recursive",
    "m2m_regress",
    "m2m_signals",
    "m2m_through",
    "m2m_through_regress",
    "m2o_recursive",
    "managers_regress",
    "many_to_many",
    "many_to_one",
    "many_to_one_null",
    "max_lengths",
    "messages_tests",
    "migrate_signals",
    "migration_test_data_persistence",
    "migrations",
    "model_fields",
    "model_forms",
    "model_formsets",
    "model_formsets_regress",
    "model_indexes",
    "model_inheritance",
    "model_inheritance_regress",
    "model_options",
    "model_package",
    "model_regress",
    "model_utils",
    "modeladmin",
    "multiple_database",
    "mutually_referential",
    "nested_foreign_keys",
    "null_fk",
    "null_fk_ordering",
    "null_queries",
    "one_to_one",
    "or_lookups",
    "order_with_respect_to",
    "ordering",
    "pagination",
    "prefetch_related",
    "proxy_model_inheritance",
    "proxy_models",
    "queries",
    "queryset_pickle",
    "redirects_tests",
    "reserved_names",
    "reverse_lookup",
    "save_delete_hooks",
    "schema",
    "select_for_update",
    "select_related",
    "select_related_onetoone",
    "select_related_regress",
    "serializers",
    "servers",
    "sessions_tests",
    "shortcuts",
    "signals",
    "sitemaps_tests",
    "sites_framework",
    "sites_tests",
    "string_lookup",
    "swappable_models",
    "syndication_tests",
    "test_client",
    "test_client_regress",
    "test_runner",
    "test_utils",
    "timezones",
    "transactions",
    "unmanaged_models",
    "update",
    "update_only_fields",
    "user_commands",
    "validation",
    "view_tests",
    "xor_lookups",
    # Add directories in django_mongodb_backend/tests
    *sorted(
        [
            x.name
            for x in (pathlib.Path(__file__).parent.parent.parent.resolve() / "tests").iterdir()
        ]
    ),
]
runtests = pathlib.Path(__file__).parent.resolve() / "runtests.py"
run_tests_cmd = f"python3 {runtests} %s --settings mongodb_settings -v 2"

shouldFail = False
for app_name in test_apps:
    res = os.system(run_tests_cmd % app_name)  # noqa: S605
    if res != 0:
        shouldFail = True
sys.exit(1 if shouldFail else 0)



from collections import defaultdict

from django.db import NotSupportedError
from django.db.models import Index
from django.db.models.lookups import BuiltinLookup
from django.db.models.sql.query import Query
from django.db.models.sql.where import AND, XOR, WhereNode
from pymongo import ASCENDING, DESCENDING
from pymongo.operations import IndexModel

from .query_utils import process_rhs

MONGO_INDEX_OPERATORS = {
    "exact": "$eq",
    "gt": "$gt",
    "gte": "$gte",
    "lt": "$lt",
    "lte": "$lte",
    "in": "$in",
}


def _get_condition_mql(self, model, schema_editor):
    """Analogous to Index._get_condition_sql()."""
    query = Query(model=model, alias_cols=False)
    where = query.build_where(self.condition)
    compiler = query.get_compiler(connection=schema_editor.connection)
    return where.as_mql_idx(compiler, schema_editor.connection)


def builtin_lookup_idx(self, compiler, connection):
    lhs_mql = self.lhs.target.column
    value = process_rhs(self, compiler, connection)
    try:
        operator = MONGO_INDEX_OPERATORS[self.lookup_name]
    except KeyError:
        raise NotSupportedError(
            f"MongoDB does not support the '{self.lookup_name}' lookup in indexes."
        ) from None
    return {lhs_mql: {operator: value}}


def get_pymongo_index_model(self, model, schema_editor, field=None, unique=False, column_prefix=""):
    """Return a pymongo IndexModel for this Django Index."""
    if self.contains_expressions:
        return None
    kwargs = {}
    filter_expression = defaultdict(dict)
    if self.condition:
        filter_expression.update(self._get_condition_mql(model, schema_editor))
    if unique:
        kwargs["unique"] = True
        # Indexing on $type matches the value of most SQL databases by
        # allowing multiple null values for the unique constraint.
        if field:
            column = column_prefix + field.column
            filter_expression[column].update({"$type": field.db_type(schema_editor.connection)})
        else:
            for field_name, _ in self.fields_orders:
                field_ = model._meta.get_field(field_name)
                filter_expression[field_.column].update(
                    {"$type": field_.db_type(schema_editor.connection)}
                )
    if filter_expression:
        kwargs["partialFilterExpression"] = filter_expression
    index_orders = (
        [(column_prefix + field.column, ASCENDING)]
        if field
        else [
            # order is "" if ASCENDING or "DESC" if DESCENDING (see
            # django.db.models.indexes.Index.fields_orders).
            (
                column_prefix + model._meta.get_field(field_name).column,
                ASCENDING if order == "" else DESCENDING,
            )
            for field_name, order in self.fields_orders
        ]
    )
    return IndexModel(index_orders, name=self.name, **kwargs)


def where_node_idx(self, compiler, connection):
    if self.connector == AND:
        operator = "$and"
    elif self.connector == XOR:
        raise NotSupportedError("MongoDB does not support the '^' operator lookup in indexes.")
    else:
        operator = "$or"
    if self.negated:
        raise NotSupportedError("MongoDB does not support the '~' operator in indexes.")
    children_mql = []
    for child in self.children:
        mql = child.as_mql_idx(compiler, connection)
        children_mql.append(mql)
    if len(children_mql) == 1:
        mql = children_mql[0]
    elif len(children_mql) > 1:
        mql = {operator: children_mql}
    else:
        mql = {}
    return mql


def register_indexes():
    BuiltinLookup.as_mql_idx = builtin_lookup_idx
    Index._get_condition_mql = _get_condition_mql
    Index.get_pymongo_index_model = get_pymongo_index_model
    WhereNode.as_mql_idx = where_node_idx



import itertools
import pprint
from collections import defaultdict

from bson import SON
from django.core.exceptions import EmptyResultSet, FieldError, FullResultSet
from django.db import IntegrityError, NotSupportedError
from django.db.models import Count
from django.db.models.aggregates import Aggregate, Variance
from django.db.models.expressions import Case, Col, OrderBy, Ref, Value, When
from django.db.models.functions.comparison import Coalesce
from django.db.models.functions.math import Power
from django.db.models.lookups import IsNull
from django.db.models.sql import compiler
from django.db.models.sql.constants import GET_ITERATOR_CHUNK_SIZE, MULTI, SINGLE
from django.db.models.sql.datastructures import BaseTable
from django.utils.functional import cached_property
from pymongo import ASCENDING, DESCENDING

from .base import Cursor
from .query import MongoQuery, wrap_database_errors


class SQLCompiler(compiler.SQLCompiler):
    """Base class for all Mongo compilers."""

    query_class = MongoQuery
    PARENT_FIELD_TEMPLATE = "parent__field__{}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aggregation_pipeline = None
        # Map columns to their subquery indices.
        self.column_indices = {}
        # A list of OrderBy objects for this query.
        self.order_by_objs = None
        self.subqueries = []

    def _get_group_alias_column(self, expr, annotation_group_idx):
        """Generate a dummy field for use in the ids fields in $group."""
        replacement = None
        if isinstance(expr, Col):
            col = expr
        else:
            # If the column is a composite expression, create a field for it.
            alias = f"__annotation_group{next(annotation_group_idx)}"
            col = self._get_column_from_expression(expr, alias)
            replacement = col
        return col.target.column, replacement

    def _get_column_from_expression(self, expr, alias):
        """
        Create a column named `alias` from the given expression to hold the
        aggregate value.
        """
        column_target = expr.output_field.clone()
        column_target.db_column = alias
        column_target.set_attributes_from_name(alias)
        return Col(self.collection_name, column_target)

    def _prepare_expressions_for_pipeline(self, expression, target, annotation_group_idx):
        """
        Prepare expressions for the aggregation pipeline.

        Handle the computation of aggregation functions used by various
        expressions. Separate and create intermediate columns, and replace
        nodes to simulate a group by operation.

        MongoDB's $group stage doesn't allow operations over the aggregator,
        e.g. COALESCE(AVG(field), 3). However, it supports operations inside
        the aggregation, e.g. AVG(number * 2).

        Handle the first case by splitting the computation into stages: compute
        the aggregation first, then apply additional operations in a subsequent
        stage by replacing the aggregate expressions with new columns prefixed
        by `__aggregation`.
        """
        replacements = {}
        group = {}
        for sub_expr in self._get_aggregate_expressions(expression):
            alias = (
                f"__aggregation{next(annotation_group_idx)}" if sub_expr != expression else target
            )
            column_target = sub_expr.output_field.clone()
            column_target.db_column = alias
            column_target.set_attributes_from_name(alias)
            inner_column = Col(self.collection_name, column_target)
            if sub_expr.distinct:
                # If the expression should return distinct values, use
                # $addToSet to deduplicate.
                rhs = sub_expr.as_mql(self, self.connection, resolve_inner_expression=True)
                group[alias] = {"$addToSet": rhs}
                replacing_expr = sub_expr.copy()
                replacing_expr.set_source_expressions([inner_column, None])
            else:
                group[alias] = sub_expr.as_mql(self, self.connection)
                replacing_expr = inner_column
            # Count must return 0 rather than null.
            if isinstance(sub_expr, Count):
                replacing_expr = Coalesce(replacing_expr, 0)
            # Variance = StdDev^2
            if isinstance(sub_expr, Variance):
                replacing_expr = Power(replacing_expr, 2)
            replacements[sub_expr] = replacing_expr
        return replacements, group

    def _prepare_annotations_for_aggregation_pipeline(self, order_by):
        """Prepare annotations for the aggregation pipeline."""
        replacements = {}
        group = {}
        annotation_group_idx = itertools.count(start=1)
        for target, expr in self.query.annotation_select.items():
            if expr.contains_aggregate:
                new_replacements, expr_group = self._prepare_expressions_for_pipeline(
                    expr, target, annotation_group_idx
                )
                replacements.update(new_replacements)
                group.update(expr_group)
        for expr, _ in order_by:
            if expr.contains_aggregate:
                new_replacements, expr_group = self._prepare_expressions_for_pipeline(
                    expr, None, annotation_group_idx
                )
                replacements.update(new_replacements)
                group.update(expr_group)
        having_replacements, having_group = self._prepare_expressions_for_pipeline(
            self.having, None, annotation_group_idx
        )
        replacements.update(having_replacements)
        group.update(having_group)
        return group, replacements

    def _get_group_expressions(self, order_by):
        if self.query.group_by is None:
            return []
        seen = set()
        expressions = set()
        if self.query.group_by is not True:
            # If group_by isn't True, then it's a list of expressions.
            for expr in self.query.group_by:
                if not hasattr(expr, "as_sql"):
                    expr = self.query.resolve_ref(expr)
                if isinstance(expr, Ref):
                    if expr.refs not in seen:
                        seen.add(expr.refs)
                        expressions.add(expr.source)
                else:
                    expressions.add(expr)
        for expr, _, alias in self.select:
            # Skip members that are already grouped.
            if alias not in seen:
                expressions |= set(expr.get_group_by_cols())
        if not self._meta_ordering:
            for expr, (_, _, is_ref) in order_by:
                # Skip references.
                if not is_ref:
                    expressions |= set(expr.get_group_by_cols())
        having_group_by = self.having.get_group_by_cols() if self.having else ()
        for expr in having_group_by:
            expressions.add(expr)
        return expressions

    def _get_group_id_expressions(self, order_by):
        """Generate group ID expressions for the aggregation pipeline."""
        replacements = {}
        group_expressions = self._get_group_expressions(order_by)
        if not group_expressions:
            ids = None
        else:
            annotation_group_idx = itertools.count(start=1)
            ids = {}
            columns = []
            for col in group_expressions:
                alias, replacement = self._get_group_alias_column(col, annotation_group_idx)
                columns.append((alias, col))
                if replacement is not None:
                    replacements[col] = replacement
                    if isinstance(col, Ref):
                        replacements[col.source] = replacement
            ids = self.get_project_fields(tuple(columns), force_expression=True)
        return ids, replacements

    def _build_aggregation_pipeline(self, ids, group):
        """Build the aggregation pipeline for grouping."""
        pipeline = []
        if not ids:
            group["_id"] = None
            pipeline.append({"$facet": {"group": [{"$group": group}]}})
            pipeline.append(
                {
                    "$addFields": {
                        key: {
                            "$getField": {
                                "input": {"$arrayElemAt": ["$group", 0]},
                                "field": key,
                            }
                        }
                        for key in group
                    }
                }
            )
        else:
            group["_id"] = ids
            pipeline.append({"$group": group})
            projected_fields = {key: f"$_id.{key}" for key in ids}
            pipeline.append({"$addFields": projected_fields})
            if "_id" not in projected_fields:
                pipeline.append({"$unset": "_id"})
        return pipeline

    def pre_sql_setup(self, with_col_aliases=False):
        extra_select, order_by, group_by = super().pre_sql_setup(with_col_aliases=with_col_aliases)
        group, all_replacements = self._prepare_annotations_for_aggregation_pipeline(order_by)
        # query.group_by is either:
        # - None: no GROUP BY
        # - True: group by select fields
        # - a list of expressions to group by.
        if group or self.query.group_by:
            ids, replacements = self._get_group_id_expressions(order_by)
            all_replacements.update(replacements)
            pipeline = self._build_aggregation_pipeline(ids, group)
            if self.having:
                having = self.having.replace_expressions(all_replacements).as_mql(
                    self, self.connection
                )
                # Add HAVING subqueries.
                for query in self.subqueries or ():
                    pipeline.extend(query.get_pipeline())
                # Remove the added subqueries.
                self.subqueries = []
                pipeline.append({"$match": {"$expr": having}})
            self.aggregation_pipeline = pipeline
        self.annotations = {
            target: expr.replace_expressions(all_replacements)
            for target, expr in self.query.annotation_select.items()
        }
        self.order_by_objs = [expr.replace_expressions(all_replacements) for expr, _ in order_by]
        return extra_select, order_by, group_by

    def execute_sql(
        self, result_type=MULTI, chunked_fetch=False, chunk_size=GET_ITERATOR_CHUNK_SIZE
    ):
        self.pre_sql_setup()
        try:
            query = self.build_query(
                # Avoid $project (columns=None) if unneeded.
                self.columns
                if self.query.annotations or not self.query.default_cols or self.query.distinct
                else None
            )
        except EmptyResultSet:
            return iter([]) if result_type == MULTI else None

        cursor = query.get_cursor()
        if result_type == SINGLE:
            try:
                obj = cursor.next()
            except StopIteration:
                return None  # No result
            else:
                return self._make_result(obj, self.columns)
        # result_type is MULTI
        cursor.batch_size(chunk_size)
        result = self.cursor_iter(cursor, chunk_size, self.columns)
        if not chunked_fetch:
            # If using non-chunked reads, read data into memory.
            return list(result)
        return result

    def results_iter(
        self,
        results=None,
        tuple_expected=False,
        chunked_fetch=False,
        chunk_size=GET_ITERATOR_CHUNK_SIZE,
    ):
        """
        Return an iterator over the results from executing query given
        to this compiler. Called by QuerySet methods.

        This method is copied from the superclass with one modification: the
        `if tuple_expected` block is deindented so that the result of
        _make_result() (a list) is cast to tuple as needed. For SQL database
        drivers, tuple results come from cursor.fetchmany(), so the cast is
        only needed there when apply_converters() casts the tuple to a list.
        This customized method could be removed if _make_result() cast its
        return value to a tuple, but that would be more expensive since that
        cast is not always needed.
        """
        if results is None:
            # QuerySet.values() or values_list()
            results = self.execute_sql(MULTI, chunked_fetch=chunked_fetch, chunk_size=chunk_size)

        fields = [s[0] for s in self.select[0 : self.col_count]]
        converters = self.get_converters(fields)
        rows = itertools.chain.from_iterable(results)
        if converters:
            rows = self.apply_converters(rows, converters)
        if tuple_expected:
            rows = map(tuple, rows)
        return rows

    def _make_result(self, entity, columns):
        """
        Decode values for the given fields from the database entity.

        The entity is assumed to be a dict using field database column
        names as keys.
        """
        result = []
        for name, col in columns:
            column_alias = getattr(col, "alias", None)
            obj = (
                # Use the related object...
                entity.get(column_alias, {})
                # ...if this column refers to an object for select_related().
                if column_alias is not None and column_alias != self.collection_name
                else entity
            )
            result.append(obj.get(name))
        return result

    def cursor_iter(self, cursor, chunk_size, columns):
        """Yield chunks of results from cursor."""
        chunk = []
        for row in cursor:
            chunk.append(self._make_result(row, columns))
            if len(chunk) == chunk_size:
                yield chunk
                chunk = []
        yield chunk

    def check_query(self):
        """Check if the current query is supported by the database."""
        if self.query.extra:
            if any(key.startswith("_prefetch_related_") for key in self.query.extra):
                raise NotSupportedError("QuerySet.prefetch_related() is not supported on MongoDB.")
            raise NotSupportedError("QuerySet.extra() is not supported on MongoDB.")

    def build_query(self, columns=None):
        """Check if the query is supported and prepare a MongoQuery."""
        self.check_query()
        query = self.query_class(self)
        ordering_fields, sort_ordering, extra_fields = self._get_ordering()
        query.ordering = sort_ordering
        if self.query.combinator:
            if not getattr(self.connection.features, f"supports_select_{self.query.combinator}"):
                raise NotSupportedError(
                    f"{self.query.combinator} is not supported on this database backend."
                )
            query.combinator_pipeline = self.get_combinator_queries()
        else:
            if self.query.distinct:
                # If query is distinct, build a $group stage for distinct
                # fields, then set project fields based on the grouped _id.
                distinct_fields = self.get_project_fields(
                    columns, ordering_fields, force_expression=True
                )
                if not query.aggregation_pipeline:
                    query.aggregation_pipeline = []
                query.aggregation_pipeline.extend(
                    [
                        {"$group": {"_id": distinct_fields}},
                        {"$project": {key: f"$_id.{key}" for key in distinct_fields}},
                    ]
                )
            else:
                # Otherwise, project fields without grouping.
                query.project_fields = self.get_project_fields(columns, ordering_fields)
            # If columns is None, then get_project_fields() won't add
            # ordering_fields to $project. Use $addFields (extra_fields) instead.
            if columns is None:
                extra_fields += ordering_fields
            query.lookup_pipeline = self.get_lookup_pipeline()
            where = self.get_where()
            try:
                expr = where.as_mql(self, self.connection) if where else {}
            except FullResultSet:
                query.match_mql = {}
            else:
                query.match_mql = {"$expr": expr}
        if extra_fields:
            query.extra_fields = self.get_project_fields(extra_fields, force_expression=True)
        query.subqueries = self.subqueries
        return query

    @cached_property
    def columns(self):
        """
        Return a tuple of (name, expression) with the columns and annotations
        which should be loaded by the query.
        """
        select_mask = self.query.get_select_mask()
        columns = (
            self.get_default_columns(select_mask) if self.query.default_cols else self.query.select
        )
        # Populate QuerySet.select_related() data.
        related_columns = []
        if self.query.select_related:
            self.get_related_selections(related_columns, select_mask)
            if related_columns:
                related_columns, _ = zip(*related_columns, strict=True)

        annotation_idx = 1

        def project_field(column):
            nonlocal annotation_idx
            if hasattr(column, "target"):
                # column is a Col.
                target = column.target.column
            # Handle Order By columns as refs columns.
            elif isinstance(column, OrderBy) and isinstance(column.expression, Ref):
                target = column.expression.refs
            else:
                # column is a Transform in values()/values_list() that needs a
                # name for $proj.
                target = f"__annotation{annotation_idx}"
                annotation_idx += 1
            return target, column

        return (
            tuple(map(project_field, columns))
            + tuple(self.annotations.items())
            + tuple(map(project_field, related_columns))
        )

    @cached_property
    def base_table(self):
        return next(
            (
                v
                for k, v in self.query.alias_map.items()
                if isinstance(v, BaseTable) and self.query.alias_refcount[k]
            ),
            None,
        )

    @cached_property
    def collection_name(self):
        if self.base_table:
            return self.base_table.table_alias or self.base_table.table_name
        # Use a dummy collection if the query doesn't specify a table.
        # For Constraint.validate():
        # SELECT 1 WHERE EXISTS(subquery checking if a constraint is violated)
        # is translated as:
        # [{"$facet": {"__null": []}},
        #  {"$lookup": {"the subquery"}},
        #  {"$match": {"condition to check from the subquery"}}]
        query = self.query_class(self)
        # The "__null" document is a placeholder so that the subquery has
        # somewhere to return its results.
        query.aggregation_pipeline = [{"$facet": {"__null": []}}]
        self.subqueries.insert(0, query)
        return "__null"

    @cached_property
    def collection(self):
        return self.connection.get_collection(self.collection_name)

    def get_combinator_queries(self):
        parts = []
        compilers = [
            query.get_compiler(self.using, self.connection, self.elide_empty)
            for query in self.query.combined_queries
        ]
        main_query_fields, _ = zip(*self.columns, strict=True)
        for compiler_ in compilers:
            try:
                # If the columns list is limited, then all combined queries
                # must have the same columns list. Set the selects defined on
                # the query on all combined queries, if not already set.
                if not compiler_.query.values_select and self.query.values_select:
                    compiler_.query = compiler_.query.clone()
                    compiler_.query.set_values(
                        (
                            *self.query.extra_select,
                            *self.query.values_select,
                            *self.query.annotation_select,
                        )
                    )
                compiler_.pre_sql_setup()
                compiler_.column_indices = self.column_indices
                columns = compiler_.columns
                parts.append((compiler_.build_query(columns), compiler_, columns))
            except EmptyResultSet:
                # Omit the empty queryset with UNION.
                if self.query.combinator == "union":
                    continue
                raise
        # Raise EmptyResultSet if all the combinator queries are empty.
        if not parts:
            raise EmptyResultSet
        # Make the combinator's stages.
        combinator_pipeline = None
        for part, compiler_, columns in parts:
            inner_pipeline = part.get_pipeline()
            # Standardize result fields.
            fields = {}
            # When a .count() is called, the main_query_field has length 1
            # otherwise it has the same length as columns.
            for alias, (ref, expr) in zip(main_query_fields, columns, strict=False):
                if isinstance(expr, Col) and expr.alias != compiler_.collection_name:
                    fields[expr.alias] = 1
                else:
                    fields[alias] = f"${ref}" if alias != ref else 1
            inner_pipeline.append({"$project": fields})
            # Combine query with the current combinator pipeline.
            if combinator_pipeline:
                combinator_pipeline.append(
                    {
                        "$unionWith": {
                            "coll": compiler_.base_table.table_name,
                            "pipeline": inner_pipeline,
                        }
                    }
                )
            else:
                combinator_pipeline = inner_pipeline
        if not self.query.combinator_all:
            ids = defaultdict(dict)
            for alias, expr in self.columns:
                # Unfold foreign fields.
                if isinstance(expr, Col) and expr.alias != self.collection_name:
                    ids[expr.alias][expr.target.column] = expr.as_mql(self, self.connection)
                else:
                    ids[alias] = f"${alias}"
            # Convert defaultdict to dict so it doesn't appear as
            # "defaultdict(<CLASS 'dict'>, ..." in query logging.
            ids = dict(ids)
            combinator_pipeline.append({"$group": {"_id": ids}})
            projected_fields = {key: f"$_id.{key}" for key in ids}
            combinator_pipeline.append({"$addFields": projected_fields})
            if "_id" not in projected_fields:
                combinator_pipeline.append({"$unset": "_id"})
        return combinator_pipeline

    def get_lookup_pipeline(self):
        result = []
        for alias in tuple(self.query.alias_map):
            if not self.query.alias_refcount[alias] or self.collection_name == alias:
                continue
            result += self.query.alias_map[alias].as_mql(self, self.connection)
        return result

    def _get_aggregate_expressions(self, expr):
        stack = [expr]
        while stack:
            expr = stack.pop()
            if isinstance(expr, Aggregate):
                yield expr
            elif hasattr(expr, "get_source_expressions"):
                stack.extend(expr.get_source_expressions())

    def get_project_fields(self, columns=None, ordering=None, force_expression=False):
        if not columns:
            return {}
        fields = defaultdict(dict)
        for name, expr in columns + (ordering or ()):
            collection = expr.alias if isinstance(expr, Col) else None
            try:
                fields[collection][name] = (
                    1
                    # For brevity/simplicity, project {"field_name": 1}
                    # instead of {"field_name": "$field_name"}.
                    if isinstance(expr, Col) and name == expr.target.column and not force_expression
                    else expr.as_mql(self, self.connection)
                )
            except EmptyResultSet:
                empty_result_set_value = getattr(expr, "empty_result_set_value", NotImplemented)
                value = (
                    False if empty_result_set_value is NotImplemented else empty_result_set_value
                )
                fields[collection][name] = Value(value).as_mql(self, self.connection)
            except FullResultSet:
                fields[collection][name] = Value(True).as_mql(self, self.connection)
        # Annotations (stored in None) and the main collection's fields
        # should appear in the top-level of the fields dict.
        fields.update(fields.pop(None, {}))
        fields.update(fields.pop(self.collection_name, {}))
        # Convert defaultdict to dict so it doesn't appear as
        # "defaultdict(<CLASS 'dict'>, ..." in query logging.
        return dict(fields)

    def _get_ordering(self):
        """
        Process the query's OrderBy objects and return:
        - A tuple of ('field_name': Col/Expression, ...)
        - A bson.SON mapping to pass to $sort.
        - A tuple of ('field_name': Expression, ...) for expressions that need
          to be added to extra_fields.
        """
        fields = []
        sort_ordering = SON()
        extra_fields = []
        idx = itertools.count(start=1)
        for order in self.order_by_objs or []:
            if isinstance(order.expression, Col):
                field_name = order.as_mql(self, self.connection).removeprefix("$")
                fields.append((order.expression.target.column, order.expression))
            elif isinstance(order.expression, Ref):
                field_name = order.as_mql(self, self.connection).removeprefix("$")
            else:
                field_name = f"__order{next(idx)}"
                fields.append((field_name, order.expression))
            # If the expression is ordered by NULLS FIRST or NULLS LAST,
            # add a field for sorting that's 1 if null or 0 if not.
            if order.nulls_first or order.nulls_last:
                null_fieldname = f"__order{next(idx)}"
                condition = When(IsNull(order.expression, True), then=Value(1))
                extra_fields.append((null_fieldname, Case(condition, default=Value(0))))
                sort_ordering[null_fieldname] = DESCENDING if order.nulls_first else ASCENDING
            sort_ordering[field_name] = DESCENDING if order.descending else ASCENDING
        return tuple(fields), sort_ordering, tuple(extra_fields)

    def get_where(self):
        return getattr(self, "where", self.query.where)

    def explain_query(self):
        # Validate format (none supported) and options.
        options = self.connection.ops.explain_query_prefix(
            self.query.explain_info.format,
            **self.query.explain_info.options,
        )
        # Build the query pipeline.
        self.pre_sql_setup()
        query = self.build_query(
            # Avoid $project (columns=None) if unneeded.
            self.columns if self.query.annotations or not self.query.default_cols else None
        )
        pipeline = query.get_pipeline()
        # Explain the pipeline.
        kwargs = {}
        for option in self.connection.ops.explain_options:
            if value := options.get(option):
                kwargs[option] = value
        explain = self.connection.database.command(
            "explain",
            {"aggregate": self.collection_name, "pipeline": pipeline, "cursor": {}},
            **kwargs,
        )
        # Generate the output: a list of lines that Django joins with newlines.
        result = []
        for key, value in explain.items():
            formatted_value = pprint.pformat(value, indent=4)
            result.append(f"{key}: {formatted_value}")
        return result


class SQLInsertCompiler(SQLCompiler):
    def execute_sql(self, returning_fields=None):
        self.pre_sql_setup()
        objs = []
        for obj in self.query.objs:
            field_values = {}
            for field in self.query.fields:
                value = field.get_db_prep_save(
                    getattr(obj, field.attname)
                    if self.query.raw
                    else field.pre_save(obj, obj._state.adding),
                    connection=self.connection,
                )
                if value is None and not field.null and not field.primary_key:
                    raise IntegrityError(
                        "You can't set %s (a non-nullable field) to None." % field.name
                    )

                field_values[field.column] = value
            objs.append(field_values)
        return self.insert(objs, returning_fields=returning_fields)

    @wrap_database_errors
    def insert(self, docs, returning_fields=None):
        """Store a list of documents using field columns as element names."""
        inserted_ids = self.collection.insert_many(docs).inserted_ids
        return [(x,) for x in inserted_ids] if returning_fields else []

    @cached_property
    def collection_name(self):
        return self.query.get_meta().db_table


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    def execute_sql(self, result_type=MULTI):
        cursor = Cursor()
        try:
            query = self.build_query()
        except EmptyResultSet:
            rowcount = 0
        else:
            rowcount = query.delete()
        cursor.rowcount = rowcount
        return cursor

    def check_query(self):
        super().check_query()
        if not self.single_alias:
            raise NotSupportedError(
                "Cannot use QuerySet.delete() when querying across multiple collections on MongoDB."
            )

    def get_where(self):
        return self.query.where

    @cached_property
    def collection_name(self):
        return self.query.base_table


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    def execute_sql(self, result_type):
        """
        Execute the specified update. Return the number of rows affected by
        the primary update query. The "primary update query" is the first
        non-empty query that is executed. Row counts for any subsequent,
        related queries are not available.
        """
        self.pre_sql_setup()
        values = {}
        for field, _, value in self.query.values:
            if hasattr(value, "resolve_expression"):
                value = value.resolve_expression(self.query, allow_joins=False, for_save=True)
                if value.contains_aggregate:
                    raise FieldError(
                        "Aggregate functions are not allowed in this query "
                        f"({field.name}={value})."
                    )
                if value.contains_over_clause:
                    raise FieldError(
                        "Window expressions are not allowed in this query "
                        f"({field.name}={value})."
                    )
            elif hasattr(value, "prepare_database_save"):
                if field.remote_field:
                    value = value.prepare_database_save(field)
                elif not hasattr(field, "embedded_model"):
                    raise TypeError(
                        f"Tried to update field {field} with a model "
                        f"instance, {value!r}. Use a value compatible with "
                        f"{field.__class__.__name__}."
                    )
            prepared = field.get_db_prep_save(value, connection=self.connection)
            if hasattr(value, "as_mql"):
                prepared = prepared.as_mql(self, self.connection)
            values[field.column] = prepared
        try:
            criteria = self.build_query().match_mql
        except EmptyResultSet:
            return 0
        is_empty = not bool(values)
        rows = 0 if is_empty else self.update(criteria, [{"$set": values}])
        for query in self.query.get_related_updates():
            aux_rows = query.get_compiler(self.using).execute_sql(result_type)
            if is_empty and aux_rows:
                rows = aux_rows
                is_empty = False
        return rows

    @wrap_database_errors
    def update(self, criteria, pipeline):
        return self.collection.update_many(criteria, pipeline).matched_count

    def check_query(self):
        super().check_query()
        if len([a for a in self.query.alias_map if self.query.alias_refcount[a]]) > 1:
            raise NotSupportedError(
                "Cannot use QuerySet.update() when querying across multiple collections on MongoDB."
            )

    def get_where(self):
        return self.query.where

    @cached_property
    def collection_name(self):
        return self.query.base_table


class SQLAggregateCompiler(SQLCompiler):
    def build_query(self, columns=None):
        query = self.query_class(self)
        query.project_fields = self.get_project_fields(tuple(self.annotations.items()))
        compiler = self.query.inner_query.get_compiler(
            self.using,
            elide_empty=self.elide_empty,
        )
        compiler.pre_sql_setup(with_col_aliases=False)
        # Avoid $project (columns=None) if unneeded.
        columns = (
            compiler.columns
            if self.query.annotations or not self.query.default_cols or self.query.distinct
            else None
        )
        subquery = compiler.build_query(columns)
        query.subqueries = [subquery]
        return query

    def _make_result(self, result, columns=None):
        return [result[k] for k in self.query.annotation_select]



import pickle
from datetime import datetime, timezone

from django.core.cache.backends.base import DEFAULT_TIMEOUT, BaseCache
from django.core.cache.backends.db import Options
from django.db import connections, router
from django.utils.functional import cached_property
from pymongo import ASCENDING, DESCENDING, IndexModel, ReturnDocument
from pymongo.errors import DuplicateKeyError, OperationFailure


class MongoSerializer:
    def __init__(self, protocol=None):
        self.protocol = pickle.HIGHEST_PROTOCOL if protocol is None else protocol

    def dumps(self, obj):
        # For better incr() and decr() atomicity, don't pickle integers.
        # Using type() rather than isinstance() matches only integers and not
        # subclasses like bool.
        if type(obj) is int:  # noqa: E721
            return obj
        return pickle.dumps(obj, self.protocol)

    def loads(self, data):
        try:
            return int(data)
        except (ValueError, TypeError):
            return pickle.loads(data)  # noqa: S301


class MongoDBCache(BaseCache):
    pickle_protocol = pickle.HIGHEST_PROTOCOL

    def __init__(self, collection_name, params):
        super().__init__(params)
        self._collection_name = collection_name

        class CacheEntry:
            _meta = Options(collection_name)

        self.cache_model_class = CacheEntry

    def create_indexes(self):
        expires_index = IndexModel("expires_at", expireAfterSeconds=0)
        key_index = IndexModel("key", unique=True)
        self.collection_for_write.create_indexes([expires_index, key_index])

    @cached_property
    def serializer(self):
        return MongoSerializer(self.pickle_protocol)

    @property
    def collection_for_read(self):
        db = router.db_for_read(self.cache_model_class)
        return connections[db].get_collection(self._collection_name)

    @property
    def collection_for_write(self):
        db = router.db_for_write(self.cache_model_class)
        return connections[db].get_collection(self._collection_name)

    def _filter_expired(self, expired=False):
        """
        Return MQL to exclude expired entries (needed because the MongoDB
        daemon does not remove expired entries precisely when they expire).
        If expired=True, return MQL to include only expired entries.
        """
        op = "$lt" if expired else "$gte"
        return {"expires_at": {op: datetime.utcnow()}}

    def get_backend_timeout(self, timeout=DEFAULT_TIMEOUT):
        if timeout is None:
            return datetime.max
        timestamp = super().get_backend_timeout(timeout)
        return datetime.fromtimestamp(timestamp, tz=timezone.utc)

    def get(self, key, default=None, version=None):
        return self.get_many([key], version).get(key, default)

    def get_many(self, keys, version=None):
        if not keys:
            return {}
        keys_map = {self.make_and_validate_key(key, version=version): key for key in keys}
        with self.collection_for_read.find(
            {"key": {"$in": tuple(keys_map)}, **self._filter_expired(expired=False)}
        ) as cursor:
            return {keys_map[row["key"]]: self.serializer.loads(row["value"]) for row in cursor}

    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_and_validate_key(key, version=version)
        num = self.collection_for_write.count_documents({}, hint="_id_")
        if num >= self._max_entries:
            self._cull(num)
        self.collection_for_write.update_one(
            {"key": key},
            {
                "$set": {
                    "key": key,
                    "value": self.serializer.dumps(value),
                    "expires_at": self.get_backend_timeout(timeout),
                }
            },
            upsert=True,
        )

    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_and_validate_key(key, version=version)
        num = self.collection_for_write.count_documents({}, hint="_id_")
        if num >= self._max_entries:
            self._cull(num)
        try:
            self.collection_for_write.update_one(
                {"key": key, **self._filter_expired(expired=True)},
                {
                    "$set": {
                        "key": key,
                        "value": self.serializer.dumps(value),
                        "expires_at": self.get_backend_timeout(timeout),
                    }
                },
                upsert=True,
            )
        except DuplicateKeyError:
            return False
        return True

    def _cull(self, num):
        if self._cull_frequency == 0:
            self.clear()
        else:
            # The fraction of entries that are culled when MAX_ENTRIES is
            # reached is 1 / CULL_FREQUENCY. For example, in the default case
            # of CULL_FREQUENCY=3, 2/3 of the entries are kept, thus `keep_num`
            # will be 2/3 of the current number of entries.
            keep_num = num - num // self._cull_frequency
            try:
                # Find the first cache entry beyond the retention limit,
                # culling entries that expire the soonest.
                deleted_from = next(
                    self.collection_for_write.aggregate(
                        [
                            {"$sort": {"expires_at": DESCENDING, "key": ASCENDING}},
                            {"$skip": keep_num},
                            {"$limit": 1},
                            {"$project": {"key": 1, "expires_at": 1}},
                        ]
                    )
                )
            except StopIteration:
                # If no entries are found, there is nothing to delete. It may
                # happen if the database removes expired entries between the
                # query to get `num` and the query to get `deleted_from`.
                pass
            else:
                # Cull the cache.
                self.collection_for_write.delete_many(
                    {
                        "$or": [
                            # Delete keys that expire before `deleted_from`...
                            {"expires_at": {"$lt": deleted_from["expires_at"]}},
                            # and the entries that share an expiration with
                            # `deleted_from` but are alphabetically after it
                            # (per the same sorting to fetch `deleted_from`).
                            {
                                "$and": [
                                    {"expires_at": deleted_from["expires_at"]},
                                    {"key": {"$gte": deleted_from["key"]}},
                                ]
                            },
                        ]
                    }
                )

    def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_and_validate_key(key, version=version)
        res = self.collection_for_write.update_one(
            {"key": key}, {"$set": {"expires_at": self.get_backend_timeout(timeout)}}
        )
        return res.matched_count > 0

    def incr(self, key, delta=1, version=None):
        serialized_key = self.make_and_validate_key(key, version=version)
        try:
            updated = self.collection_for_write.find_one_and_update(
                {"key": serialized_key, **self._filter_expired(expired=False)},
                {"$inc": {"value": delta}},
                return_document=ReturnDocument.AFTER,
            )
        except OperationFailure as exc:
            method_name = "incr" if delta >= 1 else "decr"
            raise TypeError(f"Cannot apply {method_name}() to a non-numeric value.") from exc
        if updated is None:
            raise ValueError(f"Key '{key}' not found.") from None
        return updated["value"]

    def delete(self, key, version=None):
        return self._delete_many([key], version)

    def delete_many(self, keys, version=None):
        self._delete_many(keys, version)

    def _delete_many(self, keys, version=None):
        if not keys:
            return False
        keys = tuple(self.make_and_validate_key(key, version=version) for key in keys)
        return bool(self.collection_for_write.delete_many({"key": {"$in": keys}}).deleted_count)

    def has_key(self, key, version=None):
        key = self.make_and_validate_key(key, version=version)
        num = self.collection_for_read.count_documents(
            {"key": key, **self._filter_expired(expired=False)}
        )
        return num > 0

    def clear(self):
        self.collection_for_write.delete_many({})



import datetime
import uuid
from decimal import Decimal

from bson import Decimal128
from django.db.models import Value
from django.test import SimpleTestCase


class ValueTests(SimpleTestCase):
    def test_date(self):
        self.assertEqual(
            Value(datetime.date(2025, 1, 1)).as_mql(None, None),
            datetime.datetime(2025, 1, 1),
        )

    def test_datetime(self):
        self.assertEqual(
            Value(datetime.datetime(2025, 1, 1, 9, 8, 7)).as_mql(None, None),
            datetime.datetime(2025, 1, 1, 9, 8, 7),
        )

    def test_decimal(self):
        self.assertEqual(Value(Decimal("1.0")).as_mql(None, None), Decimal128("1.0"))

    def test_time(self):
        self.assertEqual(
            Value(datetime.time(9, 8, 7)).as_mql(None, None),
            datetime.datetime(1, 1, 1, 9, 8, 7),
        )

    def test_timedelta(self):
        self.assertEqual(Value(datetime.timedelta(3600)).as_mql(None, None), 311040000000.0)

    def test_int(self):
        self.assertEqual(Value(1).as_mql(None, None), {"$literal": 1})

    def test_str(self):
        self.assertEqual(Value("foo").as_mql(None, None), "foo")

    def test_uuid(self):
        value = uuid.UUID(int=1)
        self.assertEqual(Value(value).as_mql(None, None), "00000000000000000000000000000001")



import os

from django_mongodb_backend import parse_uri

if mongodb_uri := os.getenv("MONGODB_URI"):
    db_settings = parse_uri(mongodb_uri, db_name="dummy")

    # Workaround for https://github.com/mongodb-labs/mongo-orchestration/issues/268
    if db_settings["USER"] and db_settings["PASSWORD"]:
        db_settings["OPTIONS"].update({"tls": True, "tlsAllowInvalidCertificates": True})
    DATABASES = {
        "default": {**db_settings, "NAME": "djangotests"},
        "other": {**db_settings, "NAME": "djangotests-other"},
    }
else:
    DATABASES = {
        "default": {
            "ENGINE": "django_mongodb_backend",
            "NAME": "djangotests",
        },
        "other": {
            "ENGINE": "django_mongodb_backend",
            "NAME": "djangotests-other",
        },
    }

DEFAULT_AUTO_FIELD = "django_mongodb_backend.fields.ObjectIdAutoField"
PASSWORD_HASHERS = ("django.contrib.auth.hashers.MD5PasswordHasher",)
SECRET_KEY = "django_tests_secret_key"
USE_TZ = False



from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.models import Index, UniqueConstraint

from .fields import EmbeddedModelField
from .query import wrap_database_errors
from .utils import OperationCollector


def ignore_embedded_models(func):
    """
    Make a SchemaEditor method a no-op if model is an EmbeddedModel (unless
    parent_model isn't None, in which case this is a valid recursive operation
    such as adding an index on an embedded model's field).
    """

    def wrapper(self, model, *args, **kwargs):
        parent_model = kwargs.get("parent_model")
        from .models import EmbeddedModel

        if issubclass(model, EmbeddedModel) and parent_model is None:
            return
        func(self, model, *args, **kwargs)

    return wrapper


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    def get_collection(self, name):
        if self.collect_sql:
            return OperationCollector(self.collected_sql, collection=self.connection.database[name])
        return self.connection.get_collection(name)

    def get_database(self):
        if self.collect_sql:
            return OperationCollector(self.collected_sql, db=self.connection.database)
        return self.connection.get_database()

    @wrap_database_errors
    @ignore_embedded_models
    def create_model(self, model):
        self.get_database().create_collection(model._meta.db_table)
        self._create_model_indexes(model)
        # Make implicit M2M tables.
        for field in model._meta.local_many_to_many:
            if field.remote_field.through._meta.auto_created:
                self.create_model(field.remote_field.through)

    def _create_model_indexes(self, model, column_prefix="", parent_model=None):
        """
        Create all indexes (field indexes & uniques, Meta.unique_together,
        Meta.constraints, Meta.indexes) for the model.

        If this is a recursive call due to an embedded model, `column_prefix`
        tracks the path that must be prepended to the index's column, and
        `parent_model` tracks the collection to add the index/constraint to.
        """
        if not model._meta.managed or model._meta.proxy or model._meta.swapped:
            return
        # Field indexes and uniques
        for field in model._meta.local_fields:
            if isinstance(field, EmbeddedModelField):
                new_path = f"{column_prefix}{field.column}."
                self._create_model_indexes(
                    field.embedded_model, parent_model=parent_model or model, column_prefix=new_path
                )
            if self._field_should_be_indexed(model, field):
                self._add_field_index(parent_model or model, field, column_prefix=column_prefix)
            elif self._field_should_have_unique(field):
                self._add_field_unique(parent_model or model, field, column_prefix=column_prefix)
        # Meta.unique_together
        if model._meta.unique_together:
            self.alter_unique_together(
                model,
                [],
                model._meta.unique_together,
                column_prefix=column_prefix,
                parent_model=parent_model,
            )
        # Meta.constraints
        for constraint in model._meta.constraints:
            self.add_constraint(
                model, constraint, column_prefix=column_prefix, parent_model=parent_model
            )
        # Meta.indexes
        for index in model._meta.indexes:
            self.add_index(model, index, column_prefix=column_prefix, parent_model=parent_model)

    @ignore_embedded_models
    def delete_model(self, model):
        # Delete implicit M2m tables.
        for field in model._meta.local_many_to_many:
            if field.remote_field.through._meta.auto_created:
                self.delete_model(field.remote_field.through)
        self.get_collection(model._meta.db_table).drop()

    @ignore_embedded_models
    def add_field(self, model, field):
        # Create implicit M2M tables.
        if field.many_to_many and field.remote_field.through._meta.auto_created:
            self.create_model(field.remote_field.through)
            return
        # Set default value on existing documents.
        if column := field.column:
            self.get_collection(model._meta.db_table).update_many(
                {}, [{"$set": {column: self.effective_default(field)}}]
            )
        if isinstance(field, EmbeddedModelField):
            new_path = f"{field.column}."
            self._create_model_indexes(
                field.embedded_model, parent_model=model, column_prefix=new_path
            )
        # Add an index or unique, if required.
        if self._field_should_be_indexed(model, field):
            self._add_field_index(model, field)
        elif self._field_should_have_unique(field):
            self._add_field_unique(model, field)

    @ignore_embedded_models
    def _alter_field(
        self,
        model,
        old_field,
        new_field,
        old_type,
        new_type,
        old_db_params,
        new_db_params,
        strict=False,
    ):
        collection = self.get_collection(model._meta.db_table)
        # Has unique been removed?
        old_field_unique = self._field_should_have_unique(old_field)
        new_field_unique = self._field_should_have_unique(new_field)
        if old_field_unique and not new_field_unique:
            self._remove_field_unique(model, old_field)
        # Removed an index?
        old_field_indexed = self._field_should_be_indexed(model, old_field)
        new_field_indexed = self._field_should_be_indexed(model, new_field)
        if old_field_indexed and not new_field_indexed:
            self._remove_field_index(model, old_field)
        # Have they renamed the column?
        if old_field.column != new_field.column:
            collection.update_many({}, {"$rename": {old_field.column: new_field.column}})
            # Move index to the new field, if needed.
            if old_field_indexed and new_field_indexed:
                self._remove_field_index(model, old_field)
                self._add_field_index(model, new_field)
            # Move unique to the new field, if needed.
            if old_field_unique and new_field_unique:
                self._remove_field_unique(model, old_field)
                self._add_field_unique(model, new_field)
        # Replace NULL with the field default if the field and was changed from
        # NULL to NOT NULL.
        if new_field.has_default() and old_field.null and not new_field.null:
            column = new_field.column
            default = self.effective_default(new_field)
            collection.update_many({column: {"$eq": None}}, [{"$set": {column: default}}])
        # Added an index?
        if not old_field_indexed and new_field_indexed:
            self._add_field_index(model, new_field)
        # Added a unique?
        if not old_field_unique and new_field_unique:
            self._add_field_unique(model, new_field)

    @ignore_embedded_models
    def remove_field(self, model, field):
        # Remove implicit M2M tables.
        if field.many_to_many and field.remote_field.through._meta.auto_created:
            self.delete_model(field.remote_field.through)
            return
        # Unset field on existing documents.
        if column := field.column:
            self.get_collection(model._meta.db_table).update_many({}, {"$unset": {column: ""}})
            if self._field_should_be_indexed(model, field):
                self._remove_field_index(model, field)
            elif self._field_should_have_unique(field):
                self._remove_field_unique(model, field)
        if isinstance(field, EmbeddedModelField):
            new_path = f"{field.column}."
            self._remove_model_indexes(
                field.embedded_model, parent_model=model, column_prefix=new_path
            )

    def _remove_model_indexes(self, model, column_prefix="", parent_model=None):
        """
        When removing an EmbeddedModelField, the indexes need to be removed
        recursively.
        """
        if not model._meta.managed or model._meta.proxy or model._meta.swapped:
            return
        # Field indexes and uniques
        for field in model._meta.local_fields:
            if isinstance(field, EmbeddedModelField):
                new_path = f"{column_prefix}{field.column}."
                self._remove_model_indexes(
                    field.embedded_model, parent_model=parent_model or model, column_prefix=new_path
                )
            if self._field_should_be_indexed(model, field):
                self._remove_field_index(parent_model or model, field, column_prefix=column_prefix)
            elif self._field_should_have_unique(field):
                self._remove_field_unique(parent_model or model, field, column_prefix=column_prefix)
        # Meta.unique_together
        if model._meta.unique_together:
            self.alter_unique_together(
                model,
                model._meta.unique_together,
                [],
                column_prefix=column_prefix,
                parent_model=parent_model,
            )
        # Meta.constraints
        for constraint in model._meta.constraints:
            self.remove_constraint(parent_model or model, constraint)
        # Meta.indexes
        for index in model._meta.indexes:
            self.remove_index(parent_model or model, index)

    @ignore_embedded_models
    def alter_index_together(self, model, old_index_together, new_index_together, column_prefix=""):
        olds = {tuple(fields) for fields in old_index_together}
        news = {tuple(fields) for fields in new_index_together}
        # Deleted indexes
        for field_names in olds.difference(news):
            self._remove_composed_index(
                model, field_names, {"index": True, "unique": False}, column_prefix=""
            )
        # Created indexes
        for field_names in news.difference(olds):
            self._add_composed_index(model, field_names, column_prefix=column_prefix)

    @ignore_embedded_models
    def alter_unique_together(
        self, model, old_unique_together, new_unique_together, column_prefix="", parent_model=None
    ):
        olds = {tuple(fields) for fields in old_unique_together}
        news = {tuple(fields) for fields in new_unique_together}
        # Deleted uniques
        for field_names in olds.difference(news):
            self._remove_composed_index(
                model,
                field_names,
                {"unique": True, "primary_key": False},
                column_prefix=column_prefix,
                parent_model=parent_model,
            )
        # Created uniques
        for field_names in news.difference(olds):
            columns = [model._meta.get_field(field).column for field in field_names]
            name = str(
                self._unique_constraint_name(
                    model._meta.db_table, [column_prefix + col for col in columns]
                )
            )
            constraint = UniqueConstraint(fields=field_names, name=name)
            self.add_constraint(
                model, constraint, parent_model=parent_model, column_prefix=column_prefix
            )

    @ignore_embedded_models
    def add_index(
        self, model, index, *, field=None, unique=False, column_prefix="", parent_model=None
    ):
        idx = index.get_pymongo_index_model(
            model, schema_editor=self, field=field, unique=unique, column_prefix=column_prefix
        )
        if idx:
            model = parent_model or model
            self.get_collection(model._meta.db_table).create_indexes([idx])

    def _add_composed_index(self, model, field_names, column_prefix="", parent_model=None):
        """Add an index on the given list of field_names."""
        idx = Index(fields=field_names)
        idx.set_name_with_model(model)
        self.add_index(model, idx, column_prefix=column_prefix, parent_model=parent_model)

    def _add_field_index(self, model, field, *, column_prefix=""):
        """Add an index on a field with db_index=True."""
        index = Index(fields=[column_prefix + field.name])
        index.name = self._create_index_name(model._meta.db_table, [column_prefix + field.column])
        self.add_index(model, index, field=field, column_prefix=column_prefix)

    @ignore_embedded_models
    def remove_index(self, model, index):
        if index.contains_expressions:
            return
        self.get_collection(model._meta.db_table).drop_index(index.name)

    def _remove_composed_index(
        self, model, field_names, constraint_kwargs, column_prefix="", parent_model=None
    ):
        """
        Remove the index on the given list of field_names created by
        index/unique_together, depending on constraint_kwargs.
        """
        meta_constraint_names = {constraint.name for constraint in model._meta.constraints}
        meta_index_names = {constraint.name for constraint in model._meta.indexes}
        columns = [column_prefix + model._meta.get_field(field).column for field in field_names]
        constraint_names = self._constraint_names(
            parent_model or model,
            columns,
            exclude=meta_constraint_names | meta_index_names,
            **constraint_kwargs,
        )
        if len(constraint_names) != 1:
            num_found = len(constraint_names)
            columns_str = ", ".join(columns)
            raise ValueError(
                f"Found wrong number ({num_found}) of constraints for "
                f"{model._meta.db_table}({columns_str})."
            )
        model = parent_model or model
        collection = self.get_collection(model._meta.db_table)
        collection.drop_index(constraint_names[0])

    def _remove_field_index(self, model, field, column_prefix=""):
        """Remove a field's db_index=True index."""
        collection = self.get_collection(model._meta.db_table)
        meta_index_names = {index.name for index in model._meta.indexes}
        index_names = self._constraint_names(
            model,
            [column_prefix + field.column],
            index=True,
            # Retrieve only BTREE indexes since this is what's created with
            # db_index=True.
            type_=Index.suffix,
            exclude=meta_index_names,
        )
        if len(index_names) != 1:
            num_found = len(index_names)
            raise ValueError(
                f"Found wrong number ({num_found}) of constraints for "
                f"{model._meta.db_table}.{field.column}."
            )
        collection.drop_index(index_names[0])

    @ignore_embedded_models
    def add_constraint(self, model, constraint, field=None, column_prefix="", parent_model=None):
        if isinstance(constraint, UniqueConstraint) and self._unique_supported(
            condition=constraint.condition,
            deferrable=constraint.deferrable,
            include=constraint.include,
            expressions=constraint.expressions,
            nulls_distinct=constraint.nulls_distinct,
        ):
            idx = Index(
                fields=constraint.fields,
                name=constraint.name,
                condition=constraint.condition,
            )
            self.add_index(
                model,
                idx,
                field=field,
                unique=True,
                column_prefix=column_prefix,
                parent_model=parent_model,
            )

    def _add_field_unique(self, model, field, column_prefix=""):
        name = str(
            self._unique_constraint_name(model._meta.db_table, [column_prefix + field.column])
        )
        constraint = UniqueConstraint(fields=[field.name], name=name)
        self.add_constraint(model, constraint, field=field, column_prefix=column_prefix)

    @ignore_embedded_models
    def remove_constraint(self, model, constraint):
        if isinstance(constraint, UniqueConstraint) and self._unique_supported(
            condition=constraint.condition,
            deferrable=constraint.deferrable,
            include=constraint.include,
            expressions=constraint.expressions,
            nulls_distinct=constraint.nulls_distinct,
        ):
            idx = Index(
                fields=constraint.fields,
                name=constraint.name,
                condition=constraint.condition,
            )
            self.remove_index(model, idx)

    def _remove_field_unique(self, model, field, column_prefix=""):
        # Find the unique constraint for this field
        meta_constraint_names = {constraint.name for constraint in model._meta.constraints}
        constraint_names = self._constraint_names(
            model,
            [column_prefix + field.column],
            unique=True,
            primary_key=False,
            exclude=meta_constraint_names,
        )
        if len(constraint_names) != 1:
            num_found = len(constraint_names)
            raise ValueError(
                f"Found wrong number ({num_found}) of unique constraints for "
                f"{model._meta.db_table}.{field.column}."
            )
        self.get_collection(model._meta.db_table).drop_index(constraint_names[0])

    @ignore_embedded_models
    def alter_db_table(self, model, old_db_table, new_db_table):
        if old_db_table == new_db_table:
            return
        self.get_collection(old_db_table).rename(new_db_table)

    def _field_should_have_unique(self, field):
        db_type = field.db_type(self.connection)
        # The _id column is automatically unique.
        return db_type and field.unique and field.column != "_id"



import enum

from django.db import models

from django_mongodb_backend.fields import ArrayField, EmbeddedModelField, ObjectIdField
from django_mongodb_backend.models import EmbeddedModel


# ObjectIdField
class ObjectIdModel(models.Model):
    field = ObjectIdField()


class NullableObjectIdModel(models.Model):
    field = ObjectIdField(blank=True, null=True)


class PrimaryKeyObjectIdModel(models.Model):
    field = ObjectIdField(primary_key=True)


# ArrayField
class ArrayFieldSubclass(ArrayField):
    def __init__(self, *args, **kwargs):
        super().__init__(models.IntegerField())


class Tag:
    def __init__(self, tag_id):
        self.tag_id = tag_id

    def __eq__(self, other):
        return isinstance(other, Tag) and self.tag_id == other.tag_id


class TagField(models.SmallIntegerField):
    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return Tag(int(value))

    def to_python(self, value):
        if isinstance(value, Tag):
            return value
        if value is None:
            return value
        return Tag(int(value))

    def get_prep_value(self, value):
        return value.tag_id


class IntegerArrayModel(models.Model):
    field = ArrayField(models.IntegerField(), default=list, blank=True)


class NullableIntegerArrayModel(models.Model):
    field = ArrayField(models.IntegerField(), blank=True, null=True)
    field_nested = ArrayField(ArrayField(models.IntegerField(null=True)), null=True)
    order = models.IntegerField(null=True)

    def __str__(self):
        return str(self.field)


class CharArrayModel(models.Model):
    field = ArrayField(models.CharField(max_length=10))


class DateTimeArrayModel(models.Model):
    datetimes = ArrayField(models.DateTimeField())
    dates = ArrayField(models.DateField())
    times = ArrayField(models.TimeField())


class NestedIntegerArrayModel(models.Model):
    field = ArrayField(ArrayField(models.IntegerField()))


class OtherTypesArrayModel(models.Model):
    ips = ArrayField(models.GenericIPAddressField(), default=list)
    uuids = ArrayField(models.UUIDField(), default=list)
    decimals = ArrayField(models.DecimalField(max_digits=5, decimal_places=2), default=list)
    tags = ArrayField(TagField(), blank=True, null=True)
    json = ArrayField(models.JSONField(default=dict), default=list)


class EnumField(models.CharField):
    def get_prep_value(self, value):
        return value.value if isinstance(value, enum.Enum) else value


class ArrayEnumModel(models.Model):
    array_of_enums = ArrayField(EnumField(max_length=20))


# EmbeddedModelField
class Holder(models.Model):
    data = EmbeddedModelField("Data", null=True, blank=True)


class Data(EmbeddedModel):
    integer = models.IntegerField(db_column="custom_column")
    auto_now = models.DateTimeField(auto_now=True)
    auto_now_add = models.DateTimeField(auto_now_add=True)
    json_value = models.JSONField()
    decimal = models.DecimalField(max_digits=9, decimal_places="2", null=True, blank=True)
    nested_data = EmbeddedModelField("NestedData", null=True, blank=True)


class NestedData(EmbeddedModel):
    decimal = models.DecimalField(max_digits=9, decimal_places="2", null=True, blank=True)


class Address(EmbeddedModel):
    city = models.CharField(max_length=20)
    state = models.CharField(max_length=2)
    zip_code = models.IntegerField(db_index=True)
    tags = ArrayField(models.CharField(max_length=100), null=True, blank=True)


class Author(EmbeddedModel):
    name = models.CharField(max_length=10)
    age = models.IntegerField()
    address = EmbeddedModelField(Address)
    skills = ArrayField(models.CharField(max_length=100), null=True, blank=True)


class Book(models.Model):
    name = models.CharField(max_length=100)
    author = EmbeddedModelField(Author)


class Library(models.Model):
    name = models.CharField(max_length=100)
    books = models.ManyToManyField("Book", related_name="libraries")
    best_seller = models.CharField(max_length=100, null=True, blank=True)

    def __str__(self):
        return self.name



from django.db import NotSupportedError
from django.test import SimpleTestCase

from .models import Embed


class TestMethods(SimpleTestCase):
    def test_save(self):
        e = Embed()
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be saved."):
            e.save()

    def test_delete(self):
        e = Embed()
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be deleted."):
            e.delete()


class TestManagerMethods(SimpleTestCase):
    def test_all(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be queried."):
            Embed.objects.all()

    def test_get(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be queried."):
            Embed.objects.get(foo="bar")

    def test_filter(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be queried."):
            Embed.objects.filter(foo="bar")

    def test_create(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be created."):
            Embed.objects.create(foo="bar")

    def test_update(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be updated."):
            Embed.objects.update(foo="bar")

    def test_delete(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be deleted."):
            Embed.objects.delete()

    def test_get_or_create(self):
        msg = "'EmbeddedModelManager' object has no attribute 'get_or_create'"
        with self.assertRaisesMessage(AttributeError, msg):
            Embed.objects.get_or_create()



from functools import reduce, wraps
from operator import add as add_operator

from django.core.exceptions import EmptyResultSet, FullResultSet
from django.db import DatabaseError, IntegrityError, NotSupportedError
from django.db.models.expressions import Case, Col, When
from django.db.models.functions import Mod
from django.db.models.lookups import Exact
from django.db.models.sql.constants import INNER
from django.db.models.sql.datastructures import Join
from django.db.models.sql.where import AND, OR, XOR, ExtraWhere, NothingNode, WhereNode
from pymongo.errors import BulkWriteError, DuplicateKeyError, PyMongoError


def wrap_database_errors(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BulkWriteError as e:
            if "E11000 duplicate key error" in str(e):
                raise IntegrityError from e
            raise
        except DuplicateKeyError as e:
            raise IntegrityError from e
        except PyMongoError as e:
            raise DatabaseError from e

    return wrapper


class MongoQuery:
    """
    Compilers build a MongoQuery when they want to fetch some data. They work
    by first allowing sql.compiler.SQLCompiler to partly build a sql.Query,
    constructing a MongoQuery query on top of it, and then iterating over its
    results.

    This class provides a framework for converting the SQL constraint tree
    built by Django to a "representation" more suitable for MongoDB.
    """

    def __init__(self, compiler):
        self.compiler = compiler
        self.query = compiler.query
        self.ordering = []
        self.match_mql = {}
        self.subqueries = None
        self.lookup_pipeline = None
        self.project_fields = None
        self.aggregation_pipeline = compiler.aggregation_pipeline
        self.extra_fields = None
        self.combinator_pipeline = None
        # $lookup stage that encapsulates the pipeline for performing a nested
        # subquery.
        self.subquery_lookup = None

    def __repr__(self):
        return f"<MongoQuery: {self.match_mql!r} ORDER {self.ordering!r}>"

    @wrap_database_errors
    def delete(self):
        """Execute a delete query."""
        if self.compiler.subqueries:
            raise NotSupportedError("Cannot use QuerySet.delete() when a subquery is required.")
        return self.compiler.collection.delete_many(self.match_mql).deleted_count

    @wrap_database_errors
    def get_cursor(self):
        """
        Return a pymongo CommandCursor that can be iterated on to give the
        results of the query.
        """
        return self.compiler.collection.aggregate(self.get_pipeline())

    def get_pipeline(self):
        pipeline = []
        if self.lookup_pipeline:
            pipeline.extend(self.lookup_pipeline)
        for query in self.subqueries or ():
            pipeline.extend(query.get_pipeline())
        if self.match_mql:
            pipeline.append({"$match": self.match_mql})
        if self.aggregation_pipeline:
            pipeline.extend(self.aggregation_pipeline)
        if self.project_fields:
            pipeline.append({"$project": self.project_fields})
        if self.combinator_pipeline:
            pipeline.extend(self.combinator_pipeline)
        if self.extra_fields:
            pipeline.append({"$addFields": self.extra_fields})
        if self.ordering:
            pipeline.append({"$sort": self.ordering})
        if self.query.low_mark > 0:
            pipeline.append({"$skip": self.query.low_mark})
        if self.query.high_mark is not None:
            pipeline.append({"$limit": self.query.high_mark - self.query.low_mark})
        if self.subquery_lookup:
            table_output = self.subquery_lookup["as"]
            pipeline = [
                {"$lookup": {**self.subquery_lookup, "pipeline": pipeline}},
                {
                    "$set": {
                        table_output: {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": f"${table_output}"}, "missing"]},
                                        {"$eq": [{"$size": f"${table_output}"}, 0]},
                                    ]
                                },
                                "then": {},
                                "else": {"$arrayElemAt": [f"${table_output}", 0]},
                            }
                        }
                    }
                },
            ]
        return pipeline


def extra_where(self, compiler, connection):  # noqa: ARG001
    raise NotSupportedError("QuerySet.extra() is not supported on MongoDB.")


def join(self, compiler, connection):
    lookup_pipeline = []
    lhs_fields = []
    rhs_fields = []
    # Add a join condition for each pair of joining fields.
    parent_template = "parent__field__"
    for lhs, rhs in self.join_fields:
        lhs, rhs = connection.ops.prepare_join_on_clause(
            self.parent_alias, lhs, compiler.collection_name, rhs
        )
        lhs_fields.append(lhs.as_mql(compiler, connection))
        # In the lookup stage, the reference to this column doesn't include
        # the collection name.
        rhs_fields.append(rhs.as_mql(compiler, connection))
    # Handle any join conditions besides matching field pairs.
    extra = self.join_field.get_extra_restriction(self.table_alias, self.parent_alias)
    if extra:
        columns = []
        for expr in extra.leaves():
            # Determine whether the column needs to be transformed or rerouted
            # as part of the subquery.
            for hand_side in ["lhs", "rhs"]:
                hand_side_value = getattr(expr, hand_side, None)
                if isinstance(hand_side_value, Col):
                    # If the column is not part of the joined table, add it to
                    # lhs_fields.
                    if hand_side_value.alias != self.table_alias:
                        pos = len(lhs_fields)
                        lhs_fields.append(expr.lhs.as_mql(compiler, connection))
                    else:
                        pos = None
                    columns.append((hand_side_value, pos))
        # Replace columns in the extra conditions with new column references
        # based on their rerouted positions in the join pipeline.
        replacements = {}
        for col, parent_pos in columns:
            column_target = Col(compiler.collection_name, expr.output_field.__class__())
            if parent_pos is not None:
                target_col = f"${parent_template}{parent_pos}"
                column_target.target.db_column = target_col
                column_target.target.set_attributes_from_name(target_col)
            else:
                column_target.target = col.target
            replacements[col] = column_target
        # Apply the transformed expressions in the extra condition.
        extra_condition = [extra.replace_expressions(replacements).as_mql(compiler, connection)]
    else:
        extra_condition = []

    lookup_pipeline = [
        {
            "$lookup": {
                # The right-hand table to join.
                "from": self.table_name,
                # The pipeline variables to be matched in the pipeline's
                # expression.
                "let": {
                    f"{parent_template}{i}": parent_field
                    for i, parent_field in enumerate(lhs_fields)
                },
                "pipeline": [
                    {
                        # Match the conditions:
                        #   self.table_name.field1 = parent_table.field1
                        # AND
                        #   self.table_name.field2 = parent_table.field2
                        # AND
                        #   ...
                        "$match": {
                            "$expr": {
                                "$and": [
                                    {"$eq": [f"$${parent_template}{i}", field]}
                                    for i, field in enumerate(rhs_fields)
                                ]
                                + extra_condition
                            }
                        }
                    }
                ],
                # Rename the output as table_alias.
                "as": self.table_alias,
            }
        },
    ]
    # To avoid missing data when using $unwind, an empty collection is added if
    # the join isn't an inner join. For inner joins, rows with empty arrays are
    # removed, as $unwind unrolls or unnests the array and removes the row if
    # it's empty. This is the expected behavior for inner joins. For left outer
    # joins (LOUTER), however, an empty collection is returned.
    if self.join_type != INNER:
        lookup_pipeline.append(
            {
                "$set": {
                    self.table_alias: {
                        "$cond": {
                            "if": {
                                "$or": [
                                    {"$eq": [{"$type": f"${self.table_alias}"}, "missing"]},
                                    {"$eq": [{"$size": f"${self.table_alias}"}, 0]},
                                ]
                            },
                            "then": [{}],
                            "else": f"${self.table_alias}",
                        }
                    }
                }
            }
        )
    lookup_pipeline.append({"$unwind": f"${self.table_alias}"})
    return lookup_pipeline


def where_node(self, compiler, connection):
    if self.connector == AND:
        full_needed, empty_needed = len(self.children), 1
    else:
        full_needed, empty_needed = 1, len(self.children)

    if self.connector == AND:
        operator = "$and"
    elif self.connector == XOR:
        # MongoDB doesn't support $xor, so convert:
        #   a XOR b XOR c XOR ...
        # to:
        #   (a OR b OR c OR ...) AND MOD(a + b + c + ..., 2) == 1
        # The result of an n-ary XOR is true when an odd number of operands
        # are true.
        lhs = self.__class__(self.children, OR)
        rhs_sum = reduce(
            add_operator,
            (Case(When(c, then=1), default=0) for c in self.children),
        )
        if len(self.children) > 2:
            rhs_sum = Mod(rhs_sum, 2)
        rhs = Exact(1, rhs_sum)
        return self.__class__([lhs, rhs], AND, self.negated).as_mql(compiler, connection)
    else:
        operator = "$or"

    children_mql = []
    for child in self.children:
        try:
            mql = child.as_mql(compiler, connection)
        except EmptyResultSet:
            empty_needed -= 1
        except FullResultSet:
            full_needed -= 1
        else:
            if mql:
                children_mql.append(mql)
            else:
                full_needed -= 1

        if empty_needed == 0:
            raise (FullResultSet if self.negated else EmptyResultSet)
        if full_needed == 0:
            raise (EmptyResultSet if self.negated else FullResultSet)

    if len(children_mql) == 1:
        mql = children_mql[0]
    elif len(children_mql) > 1:
        mql = {operator: children_mql}
    else:
        mql = {}

    if not mql:
        raise FullResultSet

    if self.negated and mql:
        mql = {"$not": mql}

    return mql


def register_nodes():
    ExtraWhere.as_mql = extra_where
    Join.as_mql = join
    NothingNode.as_mql = NothingNode.as_sql
    WhereNode.as_mql = where_node



import copy
from itertools import chain

from django import forms
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.utils.translation import gettext_lazy as _

from ...utils import prefix_validation_error
from ...validators import ArrayMaxLengthValidator, ArrayMinLengthValidator, LengthValidator


class SimpleArrayField(forms.CharField):
    default_error_messages = {
        "item_invalid": _("Item %(nth)s in the array did not validate:"),
    }

    def __init__(
        self, base_field, *, delimiter=",", max_length=None, min_length=None, length=None, **kwargs
    ):
        self.base_field = base_field
        self.delimiter = delimiter
        super().__init__(**kwargs)
        if (min_length is not None or max_length is not None) and length is not None:
            invalid_param = "max_length" if max_length is not None else "min_length"
            raise ImproperlyConfigured(
                f"The length and {invalid_param} parameters are mutually exclusive."
            )
        if min_length is not None:
            self.min_length = min_length
            self.validators.append(ArrayMinLengthValidator(int(min_length)))
        if max_length is not None:
            self.max_length = max_length
            self.validators.append(ArrayMaxLengthValidator(int(max_length)))
        if length is not None:
            self.length = length
            self.validators.append(LengthValidator(int(length)))

    def clean(self, value):
        value = super().clean(value)
        return [self.base_field.clean(val) for val in value]

    def prepare_value(self, value):
        if isinstance(value, list):
            return self.delimiter.join(str(self.base_field.prepare_value(v)) for v in value)
        return value

    def to_python(self, value):
        if isinstance(value, list):
            items = value
        elif value:
            items = value.split(self.delimiter)
        else:
            items = []
        errors = []
        values = []
        for index, item in enumerate(items):
            try:
                values.append(self.base_field.to_python(item))
            except ValidationError as error:
                errors.append(
                    prefix_validation_error(
                        error,
                        prefix=self.error_messages["item_invalid"],
                        code="item_invalid",
                        params={"nth": index + 1},
                    )
                )
        if errors:
            raise ValidationError(errors)
        return values

    def validate(self, value):
        super().validate(value)
        errors = []
        for index, item in enumerate(value):
            try:
                self.base_field.validate(item)
            except ValidationError as error:
                errors.append(
                    prefix_validation_error(
                        error,
                        prefix=self.error_messages["item_invalid"],
                        code="item_invalid",
                        params={"nth": index + 1},
                    )
                )
        if errors:
            raise ValidationError(errors)

    def run_validators(self, value):
        super().run_validators(value)
        errors = []
        for index, item in enumerate(value):
            try:
                self.base_field.run_validators(item)
            except ValidationError as error:
                errors.append(
                    prefix_validation_error(
                        error,
                        prefix=self.error_messages["item_invalid"],
                        code="item_invalid",
                        params={"nth": index + 1},
                    )
                )
        if errors:
            raise ValidationError(errors)

    def has_changed(self, initial, data):
        try:
            value = self.to_python(data)
        except ValidationError:
            pass
        else:
            if initial in self.empty_values and value in self.empty_values:
                return False
        return super().has_changed(initial, data)


class SplitArrayWidget(forms.Widget):
    template_name = "mongodb/widgets/split_array.html"

    def __init__(self, widget, size, **kwargs):
        self.widget = widget() if isinstance(widget, type) else widget
        self.size = size
        super().__init__(**kwargs)

    @property
    def is_hidden(self):
        return self.widget.is_hidden

    def value_from_datadict(self, data, files, name):
        return [
            self.widget.value_from_datadict(data, files, f"{name}_{index}")
            for index in range(self.size)
        ]

    def value_omitted_from_data(self, data, files, name):
        return all(
            self.widget.value_omitted_from_data(data, files, f"{name}_{index}")
            for index in range(self.size)
        )

    def id_for_label(self, id_):
        # See the comment for RadioSelect.id_for_label()
        if id_:
            id_ += "_0"
        return id_

    def get_context(self, name, value, attrs=None):
        attrs = {} if attrs is None else attrs
        context = super().get_context(name, value, attrs)
        if self.is_localized:
            self.widget.is_localized = self.is_localized
        value = value or []
        context["widget"]["subwidgets"] = []
        final_attrs = self.build_attrs(attrs)
        id_ = final_attrs.get("id")
        for i in range(max(len(value), self.size)):
            try:
                widget_value = value[i]
            except IndexError:
                widget_value = None
            if id_:
                final_attrs = {**final_attrs, "id": f"{id_}_{i}"}
            context["widget"]["subwidgets"].append(
                self.widget.get_context(name + "_%s" % i, widget_value, final_attrs)["widget"]
            )
        return context

    @property
    def media(self):
        return self.widget.media

    def __deepcopy__(self, memo):
        obj = super().__deepcopy__(memo)
        obj.widget = copy.deepcopy(self.widget)
        return obj

    @property
    def needs_multipart_form(self):
        return self.widget.needs_multipart_form


class SplitArrayField(forms.Field):
    default_error_messages = {
        "item_invalid": _("Item %(nth)s in the array did not validate:"),
    }

    def __init__(self, base_field, size, *, remove_trailing_nulls=False, **kwargs):
        self.base_field = base_field
        self.size = size
        self.remove_trailing_nulls = remove_trailing_nulls
        widget = SplitArrayWidget(widget=base_field.widget, size=size)
        kwargs.setdefault("widget", widget)
        super().__init__(**kwargs)

    def _remove_trailing_nulls(self, values):
        index = None
        if self.remove_trailing_nulls:
            for i, value in reversed(list(enumerate(values))):
                if value in self.base_field.empty_values:
                    index = i
                else:
                    break
            if index is not None:
                values = values[:index]
        return values, index

    def to_python(self, value):
        value = super().to_python(value)
        return [self.base_field.to_python(item) for item in value]

    def clean(self, value):
        cleaned_data = []
        errors = []
        if not any(value) and self.required:
            raise ValidationError(self.error_messages["required"])
        max_size = max(self.size, len(value))
        for index in range(max_size):
            item = value[index]
            try:
                cleaned_data.append(self.base_field.clean(item))
            except ValidationError as error:
                errors.append(
                    prefix_validation_error(
                        error,
                        self.error_messages["item_invalid"],
                        code="item_invalid",
                        params={"nth": index + 1},
                    )
                )
                cleaned_data.append(None)
            else:
                errors.append(None)
        cleaned_data, null_index = self._remove_trailing_nulls(cleaned_data)
        if null_index is not None:
            errors = errors[:null_index]
        errors = list(filter(None, errors))
        if errors:
            raise ValidationError(list(chain.from_iterable(errors)))
        return cleaned_data

    def has_changed(self, initial, data):
        try:
            data = self.to_python(data)
        except ValidationError:
            pass
        else:
            data, _ = self._remove_trailing_nulls(data)
            if initial in self.empty_values and data in self.empty_values:
                return False
        return super().has_changed(initial, data)



from django.core.checks import Error
from django.db import connection, models
from django.test import SimpleTestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.validation import DatabaseValidation


@isolate_apps("invalid_models_tests")
class ProhibitedFieldTests(SimpleTestCase):
    def test_autofield(self):
        class Model(models.Model):
            id = models.AutoField(primary_key=True)

        field = Model._meta.get_field("id")
        validator = DatabaseValidation(connection=connection)
        self.assertEqual(
            validator.check_field(field),
            [
                Error(
                    "MongoDB does not support AutoField.",
                    hint="Use django_mongodb_backend.fields.ObjectIdAutoField instead.",
                    obj=field,
                    id="mongodb.E001",
                )
            ],
        )

    def test_bigautofield(self):
        class Model(models.Model):
            id = models.BigAutoField(primary_key=True)

        field = Model._meta.get_field("id")
        validator = DatabaseValidation(connection=connection)
        self.assertEqual(
            validator.check_field(field),
            [
                Error(
                    "MongoDB does not support BigAutoField.",
                    hint="Use django_mongodb_backend.fields.ObjectIdAutoField instead.",
                    obj=field,
                    id="mongodb.E001",
                )
            ],
        )

    def test_smallautofield(self):
        class Model(models.Model):
            id = models.SmallAutoField(primary_key=True)

        field = Model._meta.get_field("id")
        validator = DatabaseValidation(connection=connection)
        self.assertEqual(
            validator.check_field(field),
            [
                Error(
                    "MongoDB does not support SmallAutoField.",
                    hint="Use django_mongodb_backend.fields.ObjectIdAutoField instead.",
                    obj=field,
                    id="mongodb.E001",
                )
            ],
        )



import signal

from django.db.backends.base.client import BaseDatabaseClient


class DatabaseClient(BaseDatabaseClient):
    executable_name = "mongosh"

    @classmethod
    def settings_to_cmd_args_env(cls, settings_dict, parameters):  # noqa: ARG003 parameters unused
        options = settings_dict.get("OPTIONS", {})
        args = [cls.executable_name]

        host = settings_dict["HOST"]
        port = settings_dict["PORT"]
        dbname = settings_dict["NAME"]
        user = settings_dict["USER"]
        passwd = settings_dict["PASSWORD"]
        auth_database = options.get("authSource")
        auth_mechanism = options.get("authMechanism")
        retry_writes = options.get("retryWrites")

        if host:
            args += ["--host", host]
        if port:
            args += ["--port", port]
        if user:
            args += ["--username", user]
        if passwd:
            args += ["--password", passwd]
        if auth_database:
            args += ["--authenticationDatabase", auth_database]
        if auth_mechanism:
            args += ["--authenticationMechanism", auth_mechanism]
        if retry_writes is not None:
            args += ["--retryWrites", str(retry_writes).lower()]

        args.append(dbname)

        return args, None

    def runshell(self, parameters):
        sigint_handler = signal.getsignal(signal.SIGINT)
        try:
            # Allow SIGINT to pass to mongo to abort queries.
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            super().runshell(parameters)
        finally:
            # Restore the original SIGINT handler.
            signal.signal(signal.SIGINT, sigint_handler)



from zoneinfo import ZoneInfo

from django.db import NotSupportedError
from django.db.models.functions import TruncDate, TruncTime
from django.test import TestCase, override_settings

from .models import DTModel


@override_settings(USE_TZ=True)
class TruncTests(TestCase):
    melb = ZoneInfo("Australia/Melbourne")

    def test_truncdate_tzinfo(self):
        msg = "TruncDate with tzinfo (Australia/Melbourne) isn't supported on MongoDB."
        with self.assertRaisesMessage(NotSupportedError, msg):
            DTModel.objects.annotate(
                melb_date=TruncDate("start_datetime", tzinfo=self.melb),
            ).get()

    def test_trunctime_tzinfo(self):
        msg = "TruncTime with tzinfo (Australia/Melbourne) isn't supported on MongoDB."
        with self.assertRaisesMessage(NotSupportedError, msg):
            DTModel.objects.annotate(
                melb_date=TruncTime("start_datetime", tzinfo=self.melb),
            ).get()



from datetime import datetime

from django.conf import settings
from django.db import NotSupportedError
from django.db.models import DateField, DateTimeField, TimeField
from django.db.models.expressions import Func
from django.db.models.functions.comparison import Cast, Coalesce, Greatest, Least, NullIf
from django.db.models.functions.datetime import (
    Extract,
    ExtractDay,
    ExtractHour,
    ExtractIsoWeekDay,
    ExtractIsoYear,
    ExtractMinute,
    ExtractMonth,
    ExtractSecond,
    ExtractWeek,
    ExtractWeekDay,
    ExtractYear,
    Now,
    TruncBase,
    TruncDate,
    TruncTime,
)
from django.db.models.functions.math import Ceil, Cot, Degrees, Log, Power, Radians, Random, Round
from django.db.models.functions.text import (
    Concat,
    ConcatPair,
    Left,
    Length,
    Lower,
    LTrim,
    Replace,
    RTrim,
    StrIndex,
    Substr,
    Trim,
    Upper,
)

from .query_utils import process_lhs

MONGO_OPERATORS = {
    Ceil: "ceil",
    Coalesce: "ifNull",
    Degrees: "radiansToDegrees",
    Greatest: "max",
    Least: "min",
    Power: "pow",
    Radians: "degreesToRadians",
    Random: "rand",
}
EXTRACT_OPERATORS = {
    ExtractDay.lookup_name: "dayOfMonth",
    ExtractHour.lookup_name: "hour",
    ExtractIsoWeekDay.lookup_name: "isoDayOfWeek",
    ExtractIsoYear.lookup_name: "isoWeekYear",
    ExtractMinute.lookup_name: "minute",
    ExtractMonth.lookup_name: "month",
    ExtractSecond.lookup_name: "second",
    ExtractWeek.lookup_name: "isoWeek",
    ExtractWeekDay.lookup_name: "dayOfWeek",
    ExtractYear.lookup_name: "year",
}


def cast(self, compiler, connection):
    output_type = connection.data_types[self.output_field.get_internal_type()]
    lhs_mql = process_lhs(self, compiler, connection)[0]
    if max_length := self.output_field.max_length:
        lhs_mql = {"$substrCP": [lhs_mql, 0, max_length]}
    # Skip the conversion for "object" as it doesn't need to be transformed for
    # interpretation by JSONField, which can handle types including int,
    # object, or array.
    if output_type != "object":
        lhs_mql = {"$convert": {"input": lhs_mql, "to": output_type}}
    if decimal_places := getattr(self.output_field, "decimal_places", None):
        lhs_mql = {"$trunc": [lhs_mql, decimal_places]}
    return lhs_mql


def concat(self, compiler, connection):
    return self.get_source_expressions()[0].as_mql(compiler, connection)


def concat_pair(self, compiler, connection):
    # null on either side results in null for expression, wrap with coalesce.
    coalesced = self.coalesce()
    return super(ConcatPair, coalesced).as_mql(compiler, connection)


def cot(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    return {"$divide": [1, {"$tan": lhs_mql}]}


def extract(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    operator = EXTRACT_OPERATORS.get(self.lookup_name)
    if operator is None:
        raise NotSupportedError("%s is not supported." % self.__class__.__name__)
    if timezone := self.get_tzname():
        lhs_mql = {"date": lhs_mql, "timezone": timezone}
    return {f"${operator}": lhs_mql}


def func(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    if self.function is None:
        raise NotSupportedError(f"{self} may need an as_mql() method.")
    operator = MONGO_OPERATORS.get(self.__class__, self.function.lower())
    return {f"${operator}": lhs_mql}


def left(self, compiler, connection):
    return self.get_substr().as_mql(compiler, connection)


def length(self, compiler, connection):
    # Check for null first since $strLenCP only accepts strings.
    lhs_mql = process_lhs(self, compiler, connection)
    return {"$cond": {"if": {"$eq": [lhs_mql, None]}, "then": None, "else": {"$strLenCP": lhs_mql}}}


def log(self, compiler, connection):
    # This function is usually log(base, num) but on MongoDB it's log(num, base).
    clone = self.copy()
    clone.set_source_expressions(self.get_source_expressions()[::-1])
    return func(clone, compiler, connection)


def now(self, compiler, connection):  # noqa: ARG001
    return "$$NOW"


def null_if(self, compiler, connection):
    """Return None if expr1==expr2 else expr1."""
    expr1, expr2 = (expr.as_mql(compiler, connection) for expr in self.get_source_expressions())
    return {"$cond": {"if": {"$eq": [expr1, expr2]}, "then": None, "else": expr1}}


def preserve_null(operator):
    # If the argument is null, the function should return null, not
    # $toLower/Upper's behavior of returning an empty string.
    def wrapped(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        return {
            "$cond": {
                "if": connection.mongo_operators["isnull"](lhs_mql, True),
                "then": None,
                "else": {f"${operator}": lhs_mql},
            }
        }

    return wrapped


def replace(self, compiler, connection):
    expression, text, replacement = process_lhs(self, compiler, connection)
    return {"$replaceAll": {"input": expression, "find": text, "replacement": replacement}}


def round_(self, compiler, connection):
    # Round needs its own function because it's a special case that inherits
    # from Transform but has two arguments.
    return {"$round": [expr.as_mql(compiler, connection) for expr in self.get_source_expressions()]}


def str_index(self, compiler, connection):
    lhs = process_lhs(self, compiler, connection)
    # StrIndex should be 0-indexed (not found) but it's -1-indexed on MongoDB.
    return {"$add": [{"$indexOfCP": lhs}, 1]}


def substr(self, compiler, connection):
    lhs = process_lhs(self, compiler, connection)
    # The starting index is zero-indexed on MongoDB rather than one-indexed.
    lhs[1] = {"$add": [lhs[1], -1]}
    # If no limit is specified, use the length of the string since $substrCP
    # requires one.
    if len(lhs) == 2:
        lhs.append({"$strLenCP": lhs[0]})
    return {"$substrCP": lhs}


def trim(operator):
    def wrapped(self, compiler, connection):
        lhs = process_lhs(self, compiler, connection)
        return {f"${operator}": {"input": lhs}}

    return wrapped


def trunc(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    lhs_mql = {"date": lhs_mql, "unit": self.kind, "startOfWeek": "mon"}
    if timezone := self.get_tzname():
        lhs_mql["timezone"] = timezone
    return {"$dateTrunc": lhs_mql}


def trunc_convert_value(self, value, expression, connection):
    if connection.vendor == "mongodb":
        # A custom TruncBase.convert_value() for MongoDB.
        if value is None:
            return None
        convert_to_tz = settings.USE_TZ and self.get_tzname() != "UTC"
        if isinstance(self.output_field, DateTimeField):
            if convert_to_tz:
                # Unlike other databases, MongoDB returns the value in UTC,
                # so rather than setting the time zone equal to self.tzinfo,
                # the value must be converted to tzinfo.
                value = value.astimezone(self.tzinfo)
        elif isinstance(value, datetime):
            if isinstance(self.output_field, DateField):
                if convert_to_tz:
                    value = value.astimezone(self.tzinfo)
                # Truncate for Trunc(..., output_field=DateField)
                value = value.date()
            elif isinstance(self.output_field, TimeField):
                if convert_to_tz:
                    value = value.astimezone(self.tzinfo)
                # Truncate for Trunc(..., output_field=TimeField)
                value = value.time()
        return value
    return self.convert_value(value, expression, connection)


def trunc_date(self, compiler, connection):
    # Cast to date rather than truncate to date.
    lhs_mql = process_lhs(self, compiler, connection)
    tzname = self.get_tzname()
    if tzname and tzname != "UTC":
        raise NotSupportedError(f"TruncDate with tzinfo ({tzname}) isn't supported on MongoDB.")
    return {
        "$dateFromString": {
            "dateString": {
                "$concat": [
                    {"$dateToString": {"format": "%Y-%m-%d", "date": lhs_mql}},
                    # Dates are stored with time(0, 0), so by replacing any
                    # existing time component with that, the result of
                    # TruncDate can be compared to DateField.
                    "T00:00:00.000",
                ]
            },
        }
    }


def trunc_time(self, compiler, connection):
    tzname = self.get_tzname()
    if tzname and tzname != "UTC":
        raise NotSupportedError(f"TruncTime with tzinfo ({tzname}) isn't supported on MongoDB.")
    lhs_mql = process_lhs(self, compiler, connection)
    return {
        "$dateFromString": {
            "dateString": {
                "$concat": [
                    # Times are stored with date(1, 1, 1)), so by
                    # replacing any existing date component with that, the
                    # result of TruncTime can be compared to TimeField.
                    "0001-01-01T",
                    {"$dateToString": {"format": "%H:%M:%S.%L", "date": lhs_mql}},
                ]
            }
        }
    }


def register_functions():
    Cast.as_mql = cast
    Concat.as_mql = concat
    ConcatPair.as_mql = concat_pair
    Cot.as_mql = cot
    Extract.as_mql = extract
    Func.as_mql = func
    Left.as_mql = left
    Length.as_mql = length
    Log.as_mql = log
    Lower.as_mql = preserve_null("toLower")
    LTrim.as_mql = trim("ltrim")
    Now.as_mql = now
    NullIf.as_mql = null_if
    Replace.as_mql = replace
    Round.as_mql = round_
    RTrim.as_mql = trim("rtrim")
    StrIndex.as_mql = str_index
    Substr.as_mql = substr
    Trim.as_mql = trim("trim")
    TruncBase.as_mql = trunc
    TruncBase.convert_value = trunc_convert_value
    TruncDate.as_mql = trunc_date
    TruncTime.as_mql = trunc_time
    Upper.as_mql = preserve_null("toUpper")



from bson import ObjectId
from django.core.exceptions import ValidationError
from django.test import TestCase

from .models import Order, OrderItem, Tag


class ObjectIdTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.group_id_str_1 = "1" * 24
        cls.group_id_obj_1 = ObjectId(cls.group_id_str_1)
        cls.group_id_str_2 = "2" * 24
        cls.group_id_obj_2 = ObjectId(cls.group_id_str_2)

        cls.t1 = Tag.objects.create(name="t1")
        cls.t2 = Tag.objects.create(name="t2", parent=cls.t1)
        cls.t3 = Tag.objects.create(name="t3", parent=cls.t1, group_id=cls.group_id_str_1)
        cls.t4 = Tag.objects.create(name="t4", parent=cls.t3, group_id=cls.group_id_obj_2)
        cls.t5 = Tag.objects.create(name="t5", parent=cls.t3)

    def test_filter_group_id_is_null_false(self):
        """Filter objects where group_id is not null."""
        qs = Tag.objects.filter(group_id__isnull=False).order_by("name")
        self.assertSequenceEqual(qs, [self.t3, self.t4])

    def test_filter_group_id_is_null_true(self):
        """Filter objects where group_id is null."""
        qs = Tag.objects.filter(group_id__isnull=True).order_by("name")
        self.assertSequenceEqual(qs, [self.t1, self.t2, self.t5])

    def test_filter_group_id_equal_str(self):
        """Filter by group_id with a specific string value."""
        qs = Tag.objects.filter(group_id=self.group_id_str_1).order_by("name")
        self.assertSequenceEqual(qs, [self.t3])

    def test_filter_group_id_equal_obj(self):
        """Filter by group_id with a specific ObjectId value."""
        qs = Tag.objects.filter(group_id=self.group_id_obj_1).order_by("name")
        self.assertSequenceEqual(qs, [self.t3])

    def test_filter_group_id_in_str_values(self):
        """Filter by group_id with string values in a list."""
        ids = [self.group_id_str_1, self.group_id_str_2]
        qs = Tag.objects.filter(group_id__in=ids).order_by("name")
        self.assertSequenceEqual(qs, [self.t3, self.t4])

    def test_filter_group_id_in_obj_values(self):
        """Filter by group_id with ObjectId values in a list."""
        ids = [self.group_id_obj_1, self.group_id_obj_2]
        qs = Tag.objects.filter(group_id__in=ids).order_by("name")
        self.assertSequenceEqual(qs, [self.t3, self.t4])

    def test_filter_group_id_equal_subquery(self):
        """Filter by group_id using a subquery."""
        subquery = Tag.objects.filter(name="t3").values("group_id")
        qs = Tag.objects.filter(group_id__in=subquery).order_by("name")
        self.assertSequenceEqual(qs, [self.t3])

    def test_filter_group_id_in_subquery(self):
        """Filter by group_id using a subquery with multiple values."""
        subquery = Tag.objects.filter(name__in=["t3", "t4"]).values("group_id")
        qs = Tag.objects.filter(group_id__in=subquery).order_by("name")
        self.assertSequenceEqual(qs, [self.t3, self.t4])

    def test_filter_parent_by_children_values_str(self):
        """Query to select parents of children with specific string group_id."""
        child_ids = Tag.objects.filter(group_id=self.group_id_str_1).values_list("id", flat=True)
        parent_qs = Tag.objects.filter(children__id__in=child_ids).distinct().order_by("name")
        self.assertSequenceEqual(parent_qs, [self.t1])

    def test_filter_parent_by_children_values_obj(self):
        """Query to select parents of children with specific ObjectId group_id."""
        child_ids = Tag.objects.filter(group_id=self.group_id_obj_1).values_list("id", flat=True)
        parent_qs = Tag.objects.filter(children__id__in=child_ids).distinct().order_by("name")
        self.assertSequenceEqual(parent_qs, [self.t1])

    def test_filter_group_id_union_with_str(self):
        """Combine queries using union with string values."""
        qs_a = Tag.objects.filter(group_id=self.group_id_str_1)
        qs_b = Tag.objects.filter(group_id=self.group_id_str_2)
        union_qs = qs_a.union(qs_b).order_by("name")
        self.assertSequenceEqual(union_qs, [self.t3, self.t4])

    def test_filter_group_id_union_with_obj(self):
        """Combine queries using union with ObjectId values."""
        qs_a = Tag.objects.filter(group_id=self.group_id_obj_1)
        qs_b = Tag.objects.filter(group_id=self.group_id_obj_2)
        union_qs = qs_a.union(qs_b).order_by("name")
        self.assertSequenceEqual(union_qs, [self.t3, self.t4])

    def test_filter_invalid_object_id(self):
        msg = "“value1” is not a valid Object Id.'"
        with self.assertRaisesMessage(ValidationError, msg):
            Tag.objects.filter(group_id="value1")

    def test_values_in_subquery(self):
        # If a values() queryset is used, then the given values will be used
        # instead of forcing use of the relation's field.
        o1 = Order.objects.create()
        o2 = Order.objects.create()
        oi1 = OrderItem.objects.create(order=o1, status=None)
        oi1.status = oi1.pk
        oi1.save()
        OrderItem.objects.create(order=o2, status=None)
        # The query below should match o1 as it has related order_item with
        # id == status.
        self.assertSequenceEqual(
            Order.objects.filter(items__in=OrderItem.objects.values_list("status")),
            [o1],
        )



from django import forms
from django.forms.models import modelform_factory
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _


class EmbeddedModelWidget(forms.MultiWidget):
    def __init__(self, field_names, *args, **kwargs):
        self.field_names = field_names
        super().__init__(*args, **kwargs)
        # The default widget names are "_0", "_1", etc. Use the field names
        # instead since that's how they'll be rendered by the model form.
        self.widgets_names = ["-" + name for name in field_names]

    def decompress(self, value):
        if value is None:
            return []
        # Get the data from `value` (a model) for each field.
        return [getattr(value, name) for name in self.field_names]


class EmbeddedModelBoundField(forms.BoundField):
    def __init__(self, form, field, name, prefix_override=None):
        super().__init__(form, field, name)
        # prefix_override overrides the prefix in self.field.form_kwargs so
        # that nested embedded model form elements have the correct name.
        self.prefix_override = prefix_override

    def __str__(self):
        """Render the model form as the representation for this field."""
        form = self.field.model_form_cls(instance=self.value(), **self.field.form_kwargs)
        if self.prefix_override:
            form.prefix = self.prefix_override
        return mark_safe(f"{form.as_div()}")  # noqa: S308


class EmbeddedModelField(forms.MultiValueField):
    default_error_messages = {
        "invalid": _("Enter a list of values."),
        "incomplete": _("Enter all required values."),
    }

    def __init__(self, model, prefix, *args, **kwargs):
        form_kwargs = {}
        # To avoid collisions with other fields on the form, each subfield must
        # be prefixed with the name of the field.
        form_kwargs["prefix"] = prefix
        self.form_kwargs = form_kwargs
        self.model_form_cls = modelform_factory(model, fields="__all__")
        self.model_form = self.model_form_cls(**form_kwargs)
        self.field_names = list(self.model_form.fields.keys())
        fields = self.model_form.fields.values()
        widgets = [field.widget for field in fields]
        widget = EmbeddedModelWidget(self.field_names, widgets)
        super().__init__(*args, fields=fields, widget=widget, require_all_fields=False, **kwargs)

    def compress(self, data_dict):
        if not data_dict:
            return None
        values = dict(zip(self.field_names, data_dict, strict=False))
        return self.model_form._meta.model(**values)

    def get_bound_field(self, form, field_name):
        # Nested embedded model form fields need a double prefix.
        prefix_override = f"{form.prefix}-{self.model_form.prefix}" if form.prefix else None
        return EmbeddedModelBoundField(form, self, field_name, prefix_override)

    def bound_data(self, data, initial):
        if self.disabled:
            return initial
        # Transform the bound data into a model instance.
        return self.compress(data)

    def prepare_value(self, value):
        # When rendering a form with errors, nested EmbeddedModelField data
        # won't be compressed if MultiValueField.clean() raises ValidationError
        # error before compress() is called. The data must be compressed here
        # so that EmbeddedModelBoundField.value() returns a model instance
        # (rather than a list) for initializing the form in
        # EmbeddedModelBoundField.__str__().
        return self.compress(value) if isinstance(value, list) else value



from django.core.exceptions import ValidationError
from django.test import SimpleTestCase

from django_mongodb_backend.fields import ObjectIdAutoField


class MethodTests(SimpleTestCase):
    def test_deconstruct(self):
        field = ObjectIdAutoField()
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(path, "django_mongodb_backend.fields.ObjectIdAutoField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {"primary_key": True})

    def test_get_internal_type(self):
        f = ObjectIdAutoField()
        self.assertEqual(f.get_internal_type(), "ObjectIdAutoField")

    def test_to_python_invalid_value(self):
        f = ObjectIdAutoField()
        msg = "“1” is not a valid Object Id."
        with self.assertRaisesMessage(ValidationError, msg):
            f.to_python("1")



from bson import ObjectId, errors
from django.core import exceptions
from django.db.models.fields import Field
from django.utils.translation import gettext_lazy as _

from django_mongodb_backend import forms


class ObjectIdMixin:
    default_error_messages = {
        "invalid": _("“%(value)s” is not a valid Object Id."),
    }
    description = _("Object Id")

    def db_type(self, connection):
        return "objectId"

    def rel_db_type(self, connection):
        return "objectId"

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        return self.to_python(value)

    def to_python(self, value):
        if value is None:
            return value
        try:
            return ObjectId(value)
        except (errors.InvalidId, TypeError):
            raise exceptions.ValidationError(
                self.error_messages["invalid"],
                code="invalid",
                params={"value": value},
            ) from None

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": forms.ObjectIdField,
                **kwargs,
            }
        )


class ObjectIdField(ObjectIdMixin, Field):
    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("django_mongodb_backend.fields.objectid"):
            path = path.replace(
                "django_mongodb_backend.fields.objectid",
                "django_mongodb_backend.fields",
            )
        return name, path, args, kwargs

    def get_internal_type(self):
        return "ObjectIdField"



from django.apps.registry import Apps
from django.db import models

from django_mongodb_backend.fields import EmbeddedModelField
from django_mongodb_backend.models import EmbeddedModel

# These models are inserted into a separate Apps so the test runner doesn't
# migrate them.

new_apps = Apps()


class Address(EmbeddedModel):
    city = models.CharField(max_length=20)
    state = models.CharField(max_length=2)
    zip_code = models.IntegerField(db_index=True)
    uid = models.IntegerField(unique=True)

    class Meta:
        apps = new_apps


class Author(EmbeddedModel):
    name = models.CharField(max_length=10)
    age = models.IntegerField(db_index=True)
    address = EmbeddedModelField(Address)
    employee_id = models.IntegerField(unique=True)

    class Meta:
        apps = new_apps


class Book(models.Model):
    name = models.CharField(max_length=100)
    author = EmbeddedModelField(Author)

    class Meta:
        apps = new_apps



from django.core.exceptions import ValidationError
from django.test import SimpleTestCase

from django_mongodb_backend.validators import LengthValidator


class TestLengthValidator(SimpleTestCase):
    validator = LengthValidator(10)

    def test_empty(self):
        msg = "List contains 0 items, it should contain 10."
        with self.assertRaisesMessage(ValidationError, msg):
            self.validator([])

    def test_singular(self):
        msg = "List contains 1 item, it should contain 10."
        with self.assertRaisesMessage(ValidationError, msg):
            self.validator([1])

    def test_too_short(self):
        msg = "List contains 9 items, it should contain 10."
        with self.assertRaisesMessage(ValidationError, msg):
            self.validator([1, 2, 3, 4, 5, 6, 7, 8, 9])

    def test_too_long(self):
        msg = "List contains 11 items, it should contain 10."
        with self.assertRaisesMessage(ValidationError, msg):
            self.validator(list(range(11)))

    def test_valid(self):
        self.assertEqual(self.validator(list(range(10))), None)



from django.db.models.fields import AutoField
from django.utils.functional import cached_property

from .objectid import ObjectIdMixin


class ObjectIdAutoField(ObjectIdMixin, AutoField):
    def __init__(self, *args, **kwargs):
        kwargs["db_column"] = "_id"
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.db_column == "_id":
            del kwargs["db_column"]
        if path.startswith("django_mongodb_backend.fields.auto"):
            path = path.replace(
                "django_mongodb_backend.fields.auto", "django_mongodb_backend.fields"
            )
        return name, path, args, kwargs

    def get_prep_value(self, value):
        # Override to omit super() which would call AutoField/IntegerField's
        # implementation that requires value to be an integer.
        return self.to_python(value)

    def get_internal_type(self):
        return "ObjectIdAutoField"

    @cached_property
    def validators(self):
        # Avoid IntegerField validators inherited from AutoField.
        return [*self.default_validators, *self._validators]
